#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Qiming Sun <osirpt.sun@gmail.com>
#          Susi Lehtola <susi.lehtola@gmail.com>

'''
XC functional, the interface to libxc
(https://libxc.gitlab.io)
'''

import sys
import warnings
import ctypes
import math
import numpy
from functools import lru_cache
from pyscf import lib
from pyscf.dft.xc.utils import remove_dup, format_xc_code
from pyscf.dft import xc_deriv
from pyscf import __config__

_itrf = lib.load_library('libxc_itrf')
_itrf.LIBXC_is_lda.restype = ctypes.c_int
_itrf.LIBXC_is_gga.restype = ctypes.c_int
_itrf.LIBXC_is_meta_gga.restype = ctypes.c_int
_itrf.LIBXC_needs_laplacian.restype = ctypes.c_int
_itrf.LIBXC_needs_laplacian.argtypes = [ctypes.c_int]
_itrf.LIBXC_is_hybrid.restype = ctypes.c_int
_itrf.LIBXC_is_nlc.restype = ctypes.c_int
_itrf.LIBXC_is_cam_rsh.restype = ctypes.c_int
_itrf.LIBXC_max_deriv_order.restype = ctypes.c_int
_itrf.LIBXC_number_of_functionals.restype = ctypes.c_int
_itrf.LIBXC_functional_numbers.argtypes = (numpy.ctypeslib.ndpointer(dtype=numpy.intc, ndim=1, flags=("W", "C", "A")), )
_itrf.LIBXC_functional_name.argtypes = [ctypes.c_int]
_itrf.LIBXC_functional_name.restype = ctypes.c_char_p
_itrf.LIBXC_hybrid_coeff.argtypes = [ctypes.c_int]
_itrf.LIBXC_hybrid_coeff.restype = ctypes.c_double
_itrf.LIBXC_nlc_coeff.argtypes = [ctypes.c_int,ctypes.POINTER(ctypes.c_double)]
_itrf.LIBXC_rsh_coeff.argtypes = [ctypes.c_int,ctypes.POINTER(ctypes.c_double)]

_itrf.LIBXC_version.restype = ctypes.c_char_p
_itrf.LIBXC_reference.restype = ctypes.c_char_p
_itrf.LIBXC_reference_doi.restype = ctypes.c_char_p
_itrf.LIBXC_xc_reference.argtypes = [ctypes.c_int, (ctypes.c_char_p * 8)]

_itrf.xc_functional_get_number.argtypes = (ctypes.c_char_p, )
_itrf.xc_functional_get_number.restype = ctypes.c_int

def libxc_version():
    '''Returns the version of libxc'''
    return _itrf.LIBXC_version().decode("UTF-8")

def libxc_reference():
    '''Returns the reference to libxc'''
    return _itrf.LIBXC_reference().decode("UTF-8")

def libxc_reference_doi():
    '''Returns the reference to libxc'''
    return _itrf.LIBXC_reference_doi().decode("UTF-8")

__version__ = libxc_version()
__reference__ = libxc_reference()
__reference_doi__ = libxc_reference_doi()

def available_libxc_functionals():
    # Number of functionals is
    nfunc = _itrf.LIBXC_number_of_functionals()
    # Get functional numbers
    func_ids = numpy.zeros(nfunc, dtype=numpy.intc)
    _itrf.LIBXC_functional_numbers(func_ids)
    # Returned array
    return {_itrf.LIBXC_functional_name(x).decode("UTF-8").upper() : x for x in func_ids}

XC = XC_CODES = available_libxc_functionals()
PROBLEMATIC_XC = {}

def print_XC_CODES():
    '''
    Dump the built-in libxc XC_CODES along with their references in a readable format.
    '''
    xc_codes = available_libxc_functionals()
    print('XC = XC_CODES = {')
    for name, func_id in xc_codes.items():
        refs = xc_reference(func_id)
        key = f"'{name}'"
        print(f'{key:<31s}: {func_id:<3d}, # {refs[0]}')
        for r in refs[1:]:
            print(f"                                      # {r}")
    print('}')

def _xc_key_without_underscore(xc_keys):
    new_xc = []
    for key, xc_id in xc_keys.items():
        for delimiter in ('_XC_', '_X_', '_C_', '_K_'):
            if delimiter in key:
                key0, key1 = key.split(delimiter)
                new_key1 = key1.replace('_', '').replace('-', '')
                if key1 != new_key1:
                    new_xc.append((key0+delimiter+new_key1, xc_id))
                break
    return new_xc
XC_CODES.update(_xc_key_without_underscore(XC_CODES))
del (_xc_key_without_underscore)

#
# alias
#
XC_CODES.update({
    'GGA_C_BCGP'    : 'GGA_C_ACGGA',
    'LDA'           : 1 ,
    'SLATER'        : 1 ,
    'VWN3'          : 8,
    'VWNRPA'        : 8,
    'VWN5'          : 7,
    'B88'           : 106,
    'PBE0'          : 406,
    'PBE1PBE'       : 406,
    'OPTXCORR'      : '0.7344536875999693*SLATER - 0.6984752285760186*OPTX,',
    'B3LYP'         : 402,
    'B3LYPG'        : 402,  # used by Gaussian
    'B3LYP5'        : '.2*HF + .08*SLATER + .72*B88, .81*LYP + .19*VWN', # VWN5 version
    'B3P86'         : 403,
    'B3P86G'        : 403,  # used by Gaussian
    'B3P86V5'       : '.2*HF + .08*SLATER + .72*B88, .81*P86 + .19*VWN', # VWN5 version
    #'O3LYP5'        : '.1161*HF + .9262*SLATER + .8133*OPTXCORR, .81*LYP + .19*VWN5',
    #'O3LYPG'        : '.1161*HF + .9262*SLATER + .8133*OPTXCORR, .81*LYP + .19*VWNRPA',
    'O3LYP'         : 404, # in libxc == '.1161*HF + 0.071006917*SLATER + .8133*OPTX, .81*LYP + .19*VWN5', may be erroreous
    'MPW3PW'        : 'MPW3PW5',  # VWN5 version
    'MPW3PW5'       : '.2*HF + .08*SLATER + .72*MPW91, .81*PW91 + .19*VWN',
    'MPW3PWG'       : 415,  # used by Gaussian
    'MPW3LYP'       : 'MPW3LYP5',  # VWN5 version
    'MPW3LYP5'      : '.218*HF + .073*SLATER + .709*MPW91, .871*LYP + .129*VWN',
    'MPW3LYPG'      : 419,  # used by Gaussian
    'REVB3LYP'      : 'REVB3LYP5',  # VWN5 version
    'REVB3LYP5'     : '.2*HF + .13*SLATER + .67*B88, .84*LYP + .16*VWN',
    'REVB3LYPG'     : 454,  # used by Gaussian
    'X3LYP'         : 411,
    'X3LYPG'        : 411,  # used by Gaussian
    'X3LYP5'        : '.218*HF + .073*SLATER + .542385*B88 + .166615*PW91, .871*LYP + .129*VWN',
    'CAMB3LYP'      : 'HYB_GGA_XC_CAM_B3LYP',
    'CAMYBLYP'      : 'HYB_GGA_XC_CAMY_BLYP',
    'CAMYB3LYP'     : 'HYB_GGA_XC_CAMY_B3LYP',
    'B5050LYP'      : '.5*HF + .08*SLATER + .42*B88, .81*LYP + .19*VWN',
    'MPW1LYP'       : '.25*HF + .75*MPW91, LYP',
    'MPW1PBE'       : '.25*HF + .75*MPW91, PBE',
    'PBE50'         : '.5*HF + .5*PBE, PBE',
    'REVPBE0'       : '.25*HF + .75*PBE_R, PBE',
    'B1B95'         : 440,
    'TPSS0'         : '.25*HF + .75*TPSS, TPSS',
})  # noqa: E501

if getattr(__config__, 'B3LYP_WITH_VWN5', False):
    XC_CODES['B3P86' ] = 'B3P86V5'
    XC_CODES['B3LYP' ] = 'B3LYP5'
    XC_CODES['X3LYP' ] = 'X3LYP5'

XC_KEYS = set(XC_CODES.keys())

# Some XC functionals have conventional name, like M06-L means M06-L for X
# functional and M06-L for C functional, PBE mean PBE-X plus PBE-C. If the
# conventional name was placed in the XC_CODES, it may lead to recursive
# reference when parsing the xc description.  These names (as exceptions of
# XC_CODES) are listed in XC_ALIAS below and they should be treated as a
# shortcut for XC functional.
XC_ALIAS = {
    # Conventional name : name in XC_CODES
    'BLYP'              : 'B88,LYP',
    'BP86'              : 'B88,P86',
    'PW91'              : 'PW91,PW91',
    'PBE'               : 'PBE,PBE',
    'REVPBE'            : 'PBE_R,PBE',
    'PBESOL'            : 'PBE_SOL,PBE_SOL',
    'PKZB'              : 'PKZB,PKZB',
    'TPSS'              : 'TPSS,TPSS',
    'REVTPSS'           : 'REVTPSS,REVTPSS',
    'SCAN'              : 'SCAN,SCAN',
    'RSCAN'             : 'RSCAN,RSCAN',
    'R2SCAN'            : 'R2SCAN,R2SCAN',
    'SCANL'             : 'SCANL,SCANL',
    'R2SCANL'           : 'R2SCANL,R2SCANL',
    'SOGGA'             : 'SOGGA,PBE',
    'BLOC'              : 'BLOC,TPSSLOC',
    'OLYP'              : 'OPTX,LYP',
    'OPBE'              : 'OPTX,PBE',
    'RPBE'              : 'RPBE,PBE',
    'BPBE'              : 'B88,PBE',
    'MPW91'             : 'MPW91,PW91',
    'HFLYP'             : 'HF,LYP',
    'HFPW92'            : 'HF,PW_MOD',
    'SPW92'             : 'SLATER,PW_MOD',
    'SVWN'              : 'SLATER,VWN',
    'MS0'               : 'MS0,REGTPSS',
    'MS1'               : 'MS1,REGTPSS',
    'MS2'               : 'MS2,REGTPSS',
    'MS2H'              : 'MS2H,REGTPSS',
    'MVS'               : 'MVS,REGTPSS',
    'MVSH'              : 'MVSH,REGTPSS',
    'SOGGA11'           : 'SOGGA11,SOGGA11',
    'SOGGA11_X'         : 'SOGGA11_X,SOGGA11_X',
    'KT1'               : 'KT1,VWN',
    'KT2'               : 'GGA_XC_KT2',
    'KT3'               : 'GGA_XC_KT3',
    'DLDF'              : 'DLDF,DLDF',
    'GAM'               : 'GAM,GAM',
    'M06_L'             : 'M06_L,M06_L',
    'M06_SX'            : 'M06_SX,M06_SX',
    'M11_L'             : 'M11_L,M11_L',
    'MN12_L'            : 'MN12_L,MN12_L',
    'MN15_L'            : 'MN15_L,MN15_L',
    'N12'               : 'N12,N12',
    'N12_SX'            : 'N12_SX,N12_SX',
    'MN12_SX'           : 'MN12_SX,MN12_SX',
    'MN15'              : 'MN15,MN15',
    'MBEEF'             : 'MBEEF,PBE_SOL',
    'SCAN0'             : 'SCAN0,SCAN',
    'PBEOP'             : 'PBE,OP_PBE',
    'BOP'               : 'B88,OP_B88',
    # new in libxc-4.2.3
    'REVSCAN'           : 'MGGA_X_REVSCAN,MGGA_C_REVSCAN',
    'REVSCAN_VV10'      : 'MGGA_X_REVSCAN,MGGA_C_REVSCAN_VV10',
    'SCAN_VV10'         : 'MGGA_X_SCAN,MGGA_C_SCAN_VV10',
    'SCAN_RVV10'        : 'MGGA_X_SCAN,MGGA_C_SCAN_RVV10',
    'M05'               : 'HYB_MGGA_X_M05,MGGA_C_M05',
    'M06'               : 'HYB_MGGA_X_M06,MGGA_C_M06',
    'M05_2X'            : 'HYB_MGGA_X_M05_2X,MGGA_C_M05_2X',
    'M06_2X'            : 'HYB_MGGA_X_M06_2X,MGGA_C_M06_2X',
    # extra aliases
    'SOGGA11X'          : 'SOGGA11_X',
    'M06L'              : 'M06_L',
    'M11L'              : 'M11_L',
    'MN12L'             : 'MN12_L',
    'MN15L'             : 'MN15_L',
    'N12SX'             : 'N12_SX',
    'MN12SX'            : 'MN12_SX',
    'M052X'             : 'M05_2X',
    'M062X'             : 'M06_2X',
}  # noqa: E122
XC_ALIAS.update([(key.replace('-',''), XC_ALIAS[key])
                 for key in XC_ALIAS if '-' in key])

def xc_reference(xc_code):
    '''Returns the reference to the individual XC functional'''
    hyb, fn_facs = parse_xc(xc_code)
    refs = []
    c_refs = (ctypes.c_char_p * 8)()
    for xid, fac in fn_facs:
        _itrf.LIBXC_xc_reference(xid, c_refs)
        for ref in c_refs:
            if ref:
                refs.append(ref.decode("UTF-8"))
    return refs

@lru_cache(100)
def xc_type(xc_code):
    if xc_code is None:
        return None
    elif isinstance(xc_code, str):
        if '__VV10' in xc_code:
            raise RuntimeError('Deprecated notation for NLC functional.')
        hyb, fn_facs = parse_xc(xc_code)
    else:
        assert numpy.issubdtype(type(xc_code), numpy.integer)
        fn_facs = [(xc_code, 1)]  # mimic fn_facs

    if not fn_facs:
        return 'HF'
    elif all(_itrf.LIBXC_is_lda(ctypes.c_int(xid)) for xid, fac in fn_facs):
        return 'LDA'
    elif any(_itrf.LIBXC_is_meta_gga(ctypes.c_int(xid)) for xid, fac in fn_facs):
        return 'MGGA'
    else:
        # any(_itrf.LIBXC_is_gga(ctypes.c_int(xid)) for xid, fac in fn_facs)
        # include hybrid_xc
        return 'GGA'

def is_lda(xc_code):
    return xc_type(xc_code) == 'LDA'

@lru_cache(100)
def is_hybrid_xc(xc_code):
    if xc_code is None:
        return False
    elif isinstance(xc_code, str):
        if xc_code.isdigit():
            return _itrf.LIBXC_is_hybrid(ctypes.c_int(int(xc_code)))
        else:
            if 'HF' in xc_code:
                return True
            if hybrid_coeff(xc_code) != 0:
                return True
            if rsh_coeff(xc_code) != (0, 0, 0):
                return True
            return False
    elif numpy.issubdtype(type(xc_code), numpy.integer):
        return _itrf.LIBXC_is_hybrid(ctypes.c_int(xc_code))
    else:
        return any((is_hybrid_xc(x) for x in xc_code))

def is_meta_gga(xc_code):
    return xc_type(xc_code) == 'MGGA'

def is_gga(xc_code):
    return xc_type(xc_code) == 'GGA'

@lru_cache(100)
def is_nlc(xc_code):
    # identify nlc by xc_code itself if enable_nlc is None
    if isinstance(xc_code, str):
        if xc_code.isdigit():
            return _itrf.LIBXC_is_nlc(ctypes.c_int(int(xc_code)))
        else:
            fn_facs = parse_xc(xc_code)[1]
            return any(_itrf.LIBXC_is_nlc(ctypes.c_int(xid)) for xid, fac in fn_facs)
    elif numpy.issubdtype(type(xc_code), numpy.integer):
        return _itrf.LIBXC_is_nlc(ctypes.c_int(xc_code))
    else:
        return any((is_nlc(x) for x in xc_code))

def needs_laplacian(xc_code):
    return _itrf.LIBXC_needs_laplacian(xc_code) != 0

@lru_cache(100)
def max_deriv_order(xc_code):
    hyb, fn_facs = parse_xc(xc_code)
    if fn_facs:
        return min(_itrf.LIBXC_max_deriv_order(ctypes.c_int(xid)) for xid, fac in fn_facs)
    else:
        return 3

def test_deriv_order(xc_code, deriv, raise_error=False):
    support = deriv <= max_deriv_order(xc_code)
    if not support and raise_error:
        from pyscf.dft import xcfun
        msg = ('libxc library does not support derivative order %d for  %s' %
               (deriv, xc_code))
        try:
            if xcfun.test_deriv_order(xc_code, deriv, raise_error=False):
                msg += ('''
    This functional derivative is supported in the xcfun library.
    The following code can be used to change the libxc library to xcfun library:

        from pyscf.dft import xcfun
        mf._numint.libxc = xcfun
''')
            raise NotImplementedError(msg)
        except KeyError as e:
            sys.stderr.write('\n'+msg+'\n')
            sys.stderr.write('%s not found in xcfun library\n\n' % xc_code)
            raise e
    return support

@lru_cache(100)
def hybrid_coeff(xc_code, spin=0):
    '''Support recursively defining hybrid functional
    '''
    hyb, fn_facs = parse_xc(xc_code)
    hybs = [fac * _itrf.LIBXC_hybrid_coeff(ctypes.c_int(xid)) for xid, fac in fn_facs]
    return hyb[0] + sum(hybs)

@lru_cache(100)
def nlc_coeff(xc_code):
    '''Get NLC coefficients
    '''
    hyb, fn_facs = parse_xc(xc_code)
    nlc_pars = []
    nlc_tmp = (ctypes.c_double*2)()
    for xid, fac in fn_facs:
        if _itrf.LIBXC_is_nlc(ctypes.c_int(xid)):
            _itrf.LIBXC_nlc_coeff(xid, nlc_tmp)
            nlc_pars.append((tuple(nlc_tmp), fac))
    return tuple(nlc_pars)

@lru_cache(100)
def rsh_coeff(xc_code):
    '''Range-separated parameter and HF exchange components: omega, alpha, beta

    Exc_RSH = c_LR * LR_HFX + c_SR * SR_HFX + (1-c_SR) * Ex_SR + (1-c_LR) * Ex_LR + Ec
            = alpha * HFX   + beta * SR_HFX + (1-c_SR) * Ex_SR + (1-c_LR) * Ex_LR + Ec
            = alpha * LR_HFX + hyb * SR_HFX + (1-c_SR) * Ex_SR + (1-c_LR) * Ex_LR + Ec

    SR_HFX = < pi | (1-erf(-omega r_{12}))/r_{12} | iq >
    LR_HFX = < pi | erf(-omega r_{12})/r_{12} | iq >
    alpha = c_LR
    beta = c_SR - c_LR = hyb - alpha
    '''
    if xc_code is None:
        return 0, 0, 0

    check_omega = True
    if isinstance(xc_code, str) and ',' in xc_code:
        # Parse only X part for the RSH coefficients.  This is to handle
        # exceptions for C functionals such as M11.
        xc_code = format_xc_code(xc_code)
        xc_code = xc_code.split(',')[0] + ','
        if 'SR_HF' in xc_code or 'LR_HF' in xc_code or 'RSH(' in xc_code:
            check_omega = False

    (hyb, alpha, omega), fn_facs = parse_xc(xc_code)
    if omega == 0:
        # SR and LR Coulomb share the same coefficients
        # Note: this change breaks compatibility with pyscf-2.7
        assert hyb == alpha
        beta = 0.
    else:
        beta = hyb - alpha
    rsh_pars = [omega, alpha, beta]
    rsh_tmp = (ctypes.c_double*3)()
    for xid, fac in fn_facs:
        _itrf.LIBXC_rsh_coeff(xid, rsh_tmp)
        if rsh_pars[0] == 0:
            rsh_pars[0] = rsh_tmp[0]
        elif check_omega:
            # Check functional is actually a CAM functional
            if rsh_tmp[0] != 0 and not _itrf.LIBXC_is_cam_rsh(ctypes.c_int(xid)):
                raise KeyError('Libxc functional %i employs a range separation '
                               'kernel that is not supported in PySCF' % xid)
            # Check omega
            if (rsh_tmp[0] != 0 and rsh_pars[0] != rsh_tmp[0]):
                raise ValueError('Different values of omega found for RSH functionals')
        rsh_pars[1] += rsh_tmp[1] * fac
        rsh_pars[2] += rsh_tmp[2] * fac
    return tuple(rsh_pars)

def parse_xc_name(xc_name='LDA,VWN'):
    '''Convert the XC functional name to libxc library internal ID.
    '''
    fn_facs = parse_xc(xc_name)[1]
    return fn_facs[0][0], fn_facs[1][0]

@lru_cache(100)
def parse_xc(description):
    r'''Rules to input functional description:

    * The given functional description must be a one-line string.
    * The functional description is case-insensitive.
    * The functional description string has two parts, separated by ",".  The
      first part describes the exchange functional, the second is the correlation
      functional.

      - If "," was not in string, the entire string is considered as a
        compound XC functional (including both X and C functionals, such as b3lyp).
      - To input only X functional (without C functional), leave the second
        part blank. E.g. description='slater,' means pure LDA functional.
      - To neglect X functional (just apply C functional), leave the first
        part blank. E.g. description=',vwn' means pure VWN functional.
      - If compound XC functional is specified, no matter whether it is in the
        X part (the string in front of comma) or the C part (the string behind
        comma), both X and C functionals of the compound XC functional will be
        used.

    * The functional name can be placed in arbitrary order.  Two name needs to
      be separated by operators "+" or "-".  Blank spaces are ignored.
      NOTE the parser only reads operators "+" "-" "*".  / is not in support.
    * A functional name can have at most one factor.  If the factor is not
      given, it is set to 1.  Compound functional can be scaled as a unit. For
      example '0.5*b3lyp' is equivalent to
      'HF*0.1 + .04*LDA + .36*B88, .405*LYP + .095*VWN'
    * String "HF" stands for exact exchange (HF K matrix).  Putting "HF" in
      correlation functional part is the same to putting "HF" in exchange
      part.
    * String "RSH" means range-separated operator. Its format is
      RSH(omega, alpha, beta).  Another way to input RSH is to use keywords
      SR_HF and LR_HF: "SR_HF(0.1) * alpha_plus_beta" and "LR_HF(0.1) *
      alpha" where the number in parenthesis is the value of omega.
    * Be careful with the libxc convention on GGA functional, in which the LDA
      contribution has been included.

    Args:
        description : str
            A string to describe the linear combination of different XC functionals.
            The X and C functional are separated by comma like '.8*LDA+.2*B86,VWN'.
            If "HF" was appeared in the string, it stands for the exact exchange.

    Returns:
        decoded XC description, with the data structure
        (hybrid, alpha, omega), ((libxc-Id, fac), (libxc-Id, fac), ...)
    '''  # noqa: E501

    hyb = [0, 0, 0]  # hybrid, alpha, omega (== SR_HF, LR_HF, omega)
    if description is None:
        return tuple(hyb), ()
    elif numpy.issubdtype(type(description), numpy.integer):
        return tuple(hyb), ((description, 1.),)
    elif not isinstance(description, str): #isinstance(description, (tuple,list)):
        return parse_xc('%s,%s' % tuple(description))

    description = description.upper()
    if '-D3' in description or '-D4' in description:
        from pyscf.scf.dispersion import parse_dft
        description, _, _ = parse_dft(description)
        description = description.upper()

    if (description in ('B3P86', 'B3LYP', 'X3LYP') and
        not getattr(parse_xc, 'b3lyp5_warned', False) and
        not hasattr(__config__, 'B3LYP_WITH_VWN5')):
        parse_xc.b3lyp5_warned = True
        warnings.warn('Since PySCF-2.3, B3LYP (and B3P86) are changed to the VWN-RPA variant, '
                      'corresponding to the original definition by Stephens et al. (issue 1480) '
                      'and the same as the B3LYP functional in Gaussian. '
                      'To restore the VWN5 definition, you can put the setting '
                      '"B3LYP_WITH_VWN5 = True" in pyscf_conf.py')

    def assign_omega(omega, hyb_or_sr, lr=0):
        if hyb[2] == omega or omega == 0:
            hyb[0] += hyb_or_sr
            hyb[1] += lr
        elif hyb[2] == 0:
            hyb[0] += hyb_or_sr
            hyb[1] += lr
            hyb[2] = omega
        else:
            raise ValueError('Different values of omega found for RSH functionals')
    fn_facs = []

    def parse_token(token, ftype, search_xc_alias=False):
        if token:
            if token[0] == '-':
                sign = -1
                token = token[1:]
            else:
                sign = 1
            if '*' in token:
                fac, key = token.split('*')
                if fac[0].isalpha():
                    fac, key = key, fac
                fac = sign * float(fac.replace('E_', 'E-'))
            else:
                fac, key = sign, token

            if key[:3] == 'RSH':
                # RSH(alpha; beta; omega): Range-separated-hybrid functional
                # See also utils.format_xc_code
                alpha, beta, omega = [float(x) for x in key[4:-1].split(';')]
                assign_omega(omega, fac*(alpha+beta), fac*alpha)
            elif key == 'HF':
                hyb[0] += fac
                hyb[1] += fac  # also add to LR_HF
            elif 'SR_HF' in key:
                if '(' in key:
                    omega = float(key.split('(')[1].split(')')[0])
                    assign_omega(omega, fac, 0)
                else:  # Assuming this omega the same to the existing omega
                    hyb[0] += fac
            elif 'LR_HF' in key:
                if '(' in key:
                    omega = float(key.split('(')[1].split(')')[0])
                    assign_omega(omega, 0, fac)
                else:
                    hyb[1] += fac  # == alpha
            elif key.isdigit():
                fn_facs.append((int(key), fac))
            else:
                if search_xc_alias and key in XC_ALIAS:
                    x_id = XC_ALIAS[key]
                elif key in XC_CODES:
                    x_id = XC_CODES[key]
                else:
                    possible_xc_for = fpossible_dic[ftype]
                    possible_xc = XC_KEYS.intersection(possible_xc_for(key))
                    if possible_xc:
                        if len(possible_xc) > 1:
                            sys.stderr.write('Possible xc_code %s matches %s. '
                                             % (list(possible_xc), key))
                            for x_id in possible_xc:  # Prefer X functional
                                if '_X_' in x_id:
                                    break
                            else:
                                x_id = possible_xc.pop()
                            sys.stderr.write('XC parser takes %s\n' % x_id)
                            sys.stderr.write('You can add prefix to %s for a '
                                             'specific functional (e.g. GGA_X_%s, '
                                             'HYB_MGGA_X_%s)\n'
                                             % (key, key, key))
                        else:
                            x_id = possible_xc.pop()
                        x_id = XC_CODES[x_id]
                    else:
                        # Some libxc functionals may not be listed in the
                        # XC_CODES table. Query libxc directly
                        x_id = _itrf.xc_functional_get_number(ctypes.c_char_p(key.encode()))
                        if x_id == -1:
                            raise KeyError(f"LibXCFunctional: name '{key}' not found.")
                if isinstance(x_id, str):
                    hyb1, fn_facs1 = parse_xc(x_id)
                    # Recursively scale the composed functional, to support e.g. '0.5*b3lyp'
                    if hyb1[0] != 0 or hyb1[1] != 0:
                        assign_omega(hyb1[2], hyb1[0]*fac, hyb1[1]*fac)
                    fn_facs.extend([(xid, c*fac) for xid, c in fn_facs1])
                elif x_id is None:
                    raise NotImplementedError('%s functional %s' % (ftype, key))
                else:
                    fn_facs.append((x_id, fac))
    def possible_x_for(key):
        return {key,
                    'LDA_X_'+key, 'GGA_X_'+key, 'MGGA_X_'+key,
                    'HYB_GGA_X_'+key, 'HYB_MGGA_X_'+key}
    def possible_xc_for(key):
        return {key, 'LDA_XC_'+key, 'GGA_XC_'+key, 'MGGA_XC_'+key,
                    'HYB_LDA_XC_'+key, 'HYB_GGA_XC_'+key, 'HYB_MGGA_XC_'+key}
    def possible_k_for(key):
        return {key,
                    'LDA_K_'+key, 'GGA_K_'+key,}
    def possible_x_k_for(key):
        return possible_x_for(key).union(possible_k_for(key))
    def possible_c_for(key):
        return {key,
                    'LDA_C_'+key, 'GGA_C_'+key, 'MGGA_C_'+key}
    fpossible_dic = {'X': possible_x_for,
                     'C': possible_c_for,
                     'compound XC': possible_xc_for,
                     'K': possible_k_for,
                     'X or K': possible_x_k_for}

    description = format_xc_code(description)

    if '-' in description:  # To handle e.g. M06-L
        for key in _NAME_WITH_DASH:
            if key in description:
                description = description.replace(key, _NAME_WITH_DASH[key])

    if ',' in description:
        x_code, c_code = description.split(',')
        for token in x_code.replace('-', '+-').replace(';+', ';').split('+'):
            parse_token(token, 'X or K')
        for token in c_code.replace('-', '+-').replace(';+', ';').split('+'):
            parse_token(token, 'C')
    else:
        for token in description.replace('-', '+-').replace(';+', ';').split('+'):
            # dftd3 cannot be used in a custom xc description
            assert '-d3' not in token
            parse_token(token, 'compound XC', search_xc_alias=True)
    return tuple(hyb), tuple(remove_dup(fn_facs))

_NAME_WITH_DASH = {'SR-HF'    : 'SR_HF',
                   'LR-HF'    : 'LR_HF',
                   'OTPSS-D'  : 'OTPSS_D',
                   'B97-1'    : 'B97_1',
                   'B97-2'    : 'B97_2',
                   'B97-3'    : 'B97_3',
                   'B97-K'    : 'B97_K',
                   'B97-D'    : 'B97_D',
                   'HCTH-93'  : 'HCTH_93',
                   'HCTH-120' : 'HCTH_120',
                   'HCTH-147' : 'HCTH_147',
                   'HCTH-407' : 'HCTH_407',
                   'WB97X-D'  : 'WB97X_D',
                   'WB97X-V'  : 'WB97X_V',
                   'WB97M-V'  : 'WB97M_V',
                   'WB97X-D3' : 'WB97X_D3',
                   'B97M-V'   : 'B97M_V',
                   'M05-2X'   : 'M05_2X',
                   'M06-L'    : 'M06_L',
                   'M06-HF'   : 'M06_HF',
                   'M06-2X'   : 'M06_2X',
                   'M08-HX'   : 'M08_HX',
                   'M08-SO'   : 'M08_SO',
                   'M11-L'    : 'M11_L',
                   'MN12-L'   : 'MN12_L',
                   'MN15-L'   : 'MN15_L',
                   'MN12-SX'  : 'MN12_SX',
                   'N12-SX'   : 'N12_SX',
                   'LRC-WPBE' : 'LRC_WPBE',
                   'LRC-WPBEH': 'LRC_WPBEH',
                   'LC-VV10'  : 'LC_VV10',
                   'CAM-B3LYP': 'CAM_B3LYP',
                   'E-'       : 'E_'} # For scientific notation


def eval_xc(xc_code, rho, spin=0, relativity=0, deriv=1, omega=None, verbose=None):
    r'''Interface to call libxc library to evaluate XC functional, potential
    and functional derivatives.

    * The given functional xc_code must be a one-line string.
    * The functional xc_code is case-insensitive.
    * The functional xc_code string has two parts, separated by ",".  The
      first part describes the exchange functional, the second part sets the
      correlation functional.

      - If "," not appeared in string, the entire string is treated as the
        name of a compound functional (containing both the exchange and
        the correlation functional) which was declared in the functional
        aliases list. The full list of functional aliases can be obtained by
        calling the function pyscf.dft.xcfun.XC_ALIAS.keys() .

        If the string was not found in the aliased functional list, it is
        treated as X functional.

      - To input only X functional (without C functional), leave the second
        part blank. E.g. description='slater,' means a functional with LDA
        contribution only.

      - To neglect the contribution of X functional (just apply C functional),
        leave blank in the first part, e.g. description=',vwn' means a
        functional with VWN only.

      - If compound XC functional is specified, no matter whether it is in the
        X part (the string in front of comma) or the C part (the string behind
        comma), both X and C functionals of the compound XC functional will be
        used.

    * The functional name can be placed in arbitrary order.  Two names need to
      be separated by operators "+" or "-".  Blank spaces are ignored.
      NOTE the parser only reads operators "+" "-" "*".  / is not supported.

    * A functional name can have at most one factor.  If the factor is not
      given, it is set to 1.  Compound functional can be scaled as a unit. For
      example '0.5*b3lyp' is equivalent to
      'HF*0.1 + .04*LDA + .36*B88, .405*LYP + .095*VWN'

    * String "HF" stands for exact exchange (HF K matrix).  "HF" can be put in
      the correlation functional part (after comma). Putting "HF" in the
      correlation part is the same to putting "HF" in the exchange part.

    * String "RSH" means range-separated operator. Its format is
      RSH(omega, alpha, beta).  Another way to input RSH is to use keywords
      SR_HF and LR_HF: "SR_HF(0.1) * alpha_plus_beta" and "LR_HF(0.1) *
      alpha" where the number in parenthesis is the value of omega.

    * Be careful with the libxc convention of GGA functional, in which the LDA
      contribution is included.

    Args:
        xc_code : str
            A string to describe the linear combination of different XC functionals.
            The X and C functional are separated by comma like '.8*LDA+.2*B86,VWN'.
            If "HF" (exact exchange) is appeared in the string, the HF part will
            be skipped.  If an empty string "" is given, the returns exc, vxc,...
            will be vectors of zeros.
        rho : ndarray
            Shape of ((*,N)) for electron density (and derivatives) if spin = 0;
            Shape of ((*,N),(*,N)) for alpha/beta electron density (and derivatives) if spin > 0;
            where N is number of grids.
            rho (*,N) are ordered as (den,grad_x,grad_y,grad_z,laplacian,tau)
            where grad_x = d/dx den, laplacian = \nabla^2 den, tau = 1/2(\nabla f)^2
            In spin unrestricted case,
            rho is ((den_u,grad_xu,grad_yu,grad_zu,laplacian_u,tau_u)
                    (den_d,grad_xd,grad_yd,grad_zd,laplacian_d,tau_d))

    Kwargs:
        spin : int
            spin polarized if spin > 0
        relativity : int
            No effects.
        verbose : int or object of :class:`Logger`
            No effects.

    Returns:
        ex, vxc, fxc, kxc

        where

        * vxc = (vrho, vsigma, vlapl, vtau) for restricted case

        * vxc for unrestricted case
          | vrho[:,2]   = (u, d)
          | vsigma[:,3] = (uu, ud, dd)
          | vlapl[:,2]  = (u, d)
          | vtau[:,2]   = (u, d)

        * fxc for restricted case:
          (v2rho2, v2rhosigma, v2sigma2, v2lapl2, v2tau2, v2rholapl, v2rhotau, v2lapltau, v2sigmalapl, v2sigmatau)

        * fxc for unrestricted case:
          | v2rho2[:,3]     = (u_u, u_d, d_d)
          | v2rhosigma[:,6] = (u_uu, u_ud, u_dd, d_uu, d_ud, d_dd)
          | v2sigma2[:,6]   = (uu_uu, uu_ud, uu_dd, ud_ud, ud_dd, dd_dd)
          | v2lapl2[:,3]
          | v2tau2[:,3]     = (u_u, u_d, d_d)
          | v2rholapl[:,4]
          | v2rhotau[:,4]   = (u_u, u_d, d_u, d_d)
          | v2lapltau[:,4]
          | v2sigmalapl[:,6]
          | v2sigmatau[:,6] = (uu_u, uu_d, ud_u, ud_d, dd_u, dd_d)

        * kxc for restricted case:
          (v3rho3, v3rho2sigma, v3rhosigma2, v3sigma3,
           v3rho2lapl, v3rho2tau,
           v3rhosigmalapl, v3rhosigmatau,
           v3rholapl2, v3rholapltau, v3rhotau2,
           v3sigma2lapl, v3sigma2tau,
           v3sigmalapl2, v3sigmalapltau, v3sigmatau2,
           v3lapl3, v3lapl2tau, v3lapltau2, v3tau3)

        * kxc for unrestricted case:
          | v3rho3[:,4]         = (u_u_u, u_u_d, u_d_d, d_d_d)
          | v3rho2sigma[:,9]    = (u_u_uu, u_u_ud, u_u_dd, u_d_uu, u_d_ud, u_d_dd, d_d_uu, d_d_ud, d_d_dd)
          | v3rhosigma2[:,12]   = (u_uu_uu, u_uu_ud, u_uu_dd, u_ud_ud, u_ud_dd, u_dd_dd, d_uu_uu, d_uu_ud, d_uu_dd, d_ud_ud, d_ud_dd, d_dd_dd)
          | v3sigma3[:,10]      = (uu_uu_uu, uu_uu_ud, uu_uu_dd, uu_ud_ud, uu_ud_dd, uu_dd_dd, ud_ud_ud, ud_ud_dd, ud_dd_dd, dd_dd_dd)
          | v3rho2lapl[:,6]
          | v3rho2tau[:,6]      = (u_u_u, u_u_d, u_d_u, u_d_d, d_d_u, d_d_d)
          | v3rhosigmalapl[:,12]
          | v3rhosigmatau[:,12] = (u_uu_u, u_uu_d, u_ud_u, u_ud_d, u_dd_u, u_dd_d,
                                   d_uu_u, d_uu_d, d_ud_u, d_ud_d, d_dd_u, d_dd_d)
          | v3rholapl2[:,6]
          | v3rholapltau[:,8]
          | v3rhotau2[:,6]      = (u_u_u, u_u_d, u_d_d, d_u_u, d_u_d, d_d_d)
          | v3sigma2lapl[:,12]
          | v3sigma2tau[:,12]   = (uu_uu_u, uu_uu_d, uu_ud_u, uu_ud_d, uu_dd_u, uu_dd_d,
                                   ud_ud_u, ud_ud_d, ud_dd_u, ud_dd_d, dd_dd_u, dd_dd_d)
          | v3sigmalapl2[:,9]
          | v3sigmalapltau[:,12]
          | v3sigmatau2[:,9]    = (uu_u_u, uu_u_d, uu_d_d, ud_u_u, ud_u_d, ud_d_d, dd_u_u, dd_u_d, dd_d_d)
          | v3lapl3[:,4]
          | v3lapl2tau[:,6]
          | v3lapltau2[:,6]
          | v3tau3[:,4]         = (u_u_u, u_u_d, u_d_d, d_d_d)

        see also libxc_itrf.c
    '''  # noqa: E501
    outbuf = _eval_xc(xc_code, rho, spin, deriv, omega)
    exc = outbuf[0]
    vxc = fxc = kxc = None
    xctype = xc_type(xc_code)
    if xctype == 'LDA' and spin == 0:
        if deriv > 0:
            vxc = [outbuf[1]]
        if deriv > 1:
            fxc = [outbuf[2]]
        if deriv > 2:
            kxc = [outbuf[3]]
    elif xctype == 'GGA' and spin == 0:
        if deriv > 0:
            vxc = [outbuf[1], outbuf[2]]
        if deriv > 1:
            fxc = [outbuf[3], outbuf[4], outbuf[5]]
        if deriv > 2:
            kxc = [outbuf[6], outbuf[7], outbuf[8], outbuf[9]]
    elif xctype == 'LDA' and spin == 1:
        if deriv > 0:
            vxc = [outbuf[1:3].T]
        if deriv > 1:
            fxc = [outbuf[3:6].T]
        if deriv > 2:
            kxc = [outbuf[6:10].T]
    elif xctype == 'GGA' and spin == 1:
        if deriv > 0:
            vxc = [outbuf[1:3].T, outbuf[3:6].T]
        if deriv > 1:
            fxc = [outbuf[6:9].T, outbuf[9:15].T, outbuf[15:21].T]
        if deriv > 2:
            kxc = [outbuf[21:25].T, outbuf[25:34].T, outbuf[34:46].T, outbuf[46:56].T]
    elif xctype == 'MGGA' and spin == 0:
        if deriv > 0:
            vxc = [outbuf[1], outbuf[2], None, outbuf[3]]
        if deriv > 1:
            fxc = [
                # v2rho2, v2rhosigma, v2sigma2,
                outbuf[4], outbuf[5], outbuf[6],
                # v2lapl2, v2tau2,
                None, outbuf[9],
                # v2rholapl, v2rhotau,
                None, outbuf[7],
                # v2lapltau, v2sigmalapl, v2sigmatau,
                None, None, outbuf[8]]
        if deriv > 2:
            # v3lapltau2 might not be strictly 0
            # outbuf[18] = 0
            kxc = [
                # v3rho3, v3rho2sigma, v3rhosigma2, v3sigma3,
                outbuf[10], outbuf[11], outbuf[12], outbuf[13],
                # v3rho2lapl, v3rho2tau,
                None, outbuf[14],
                # v3rhosigmalapl, v3rhosigmatau,
                None, outbuf[15],
                # v3rholapl2, v3rholapltau, v3rhotau2,
                None, None, outbuf[16],
                # v3sigma2lapl, v3sigma2tau,
                None, outbuf[17],
                # v3sigmalapl2, v3sigmalapltau, v3sigmatau2,
                None, None, outbuf[18],
                # v3lapl3, v3lapl2tau, v3lapltau2, v3tau3)
                None, None, None, outbuf[19]]
    elif xctype == 'MGGA' and spin == 1:
        if deriv > 0:
            vxc = [outbuf[1:3].T, outbuf[3:6].T, None, outbuf[6:8].T]
        if deriv > 1:
            # v2lapltau might not be strictly 0
            # outbuf[39:43] = 0
            fxc = [
                # v2rho2, v2rhosigma, v2sigma2,
                outbuf[8:11].T, outbuf[11:17].T, outbuf[17:23].T,
                # v2lapl2, v2tau2,
                None, outbuf[33:36].T,
                # v2rholapl, v2rhotau,
                None, outbuf[23:27].T,
                # v2lapltau, v2sigmalapl, v2sigmatau,
                None, None, outbuf[27:33].T]
        if deriv > 2:
            # v3lapltau2 might not be strictly 0
            # outbuf[204:216] = 0
            kxc = [
                # v3rho3, v3rho2sigma, v3rhosigma2, v3sigma3,
                outbuf[36:40].T, outbuf[40:49].T, outbuf[49:61].T, outbuf[61:71].T,
                # v3rho2lapl, v3rho2tau,
                None, outbuf[71:77].T,
                # v3rhosigmalapl, v3rhosigmatau,
                None, outbuf[77:89].T,
                # v3rholapl2, v3rholapltau, v3rhotau2,
                None, None, outbuf[89:95].T,
                # v3sigma2lapl, v3sigma2tau,
                None, outbuf[95:107].T,
                # v3sigmalapl2, v3sigmalapltau, v3sigmatau2,
                None, None, outbuf[107:116].T,
                # v3lapl3, v3lapl2tau, v3lapltau2, v3tau3)
                None, None, None, outbuf[116:120].T]
    return exc, vxc, fxc, kxc

_GGA_SORT = {
    (1, 2): numpy.array([
        6, 7, 9, 10, 11, 8, 12, 13, 14, 15, 16, 17, 18, 19, 20,
    ]),
    (1, 3): numpy.array([
        21, 22, 25, 26, 27, 23, 28, 29, 30, 34, 35, 36, 37, 38, 39, 24, 31, 32,
        33, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55,
    ]),
    (1, 4): numpy.array([
        56, 57, 61, 62, 63, 58, 64, 65, 66, 73, 74, 75, 76, 77, 78, 59, 67, 68,
        69, 79, 80, 81, 82, 83, 84, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 60,
        70, 71, 72, 85, 86, 87, 88, 89, 90, 101, 102, 103, 104, 105, 106, 107,
        108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121,
        122, 123, 124, 125,
    ])
}
_MGGA_SORT = {
    (0, 2): numpy.array([
        0, # v2rho2
        1, # v2rhosigma
        3, # v2rhotau
        2, # v2sigma2
        4, # v2sigmatau
        5, # v2tau2
    ]) + 4,
    (0, 3): numpy.array([
        0, # v3rho3
        1, # v3rho2sigma
        4, # v3rho2tau
        2, # v3rhosigma2
        5, # v3rhosigmatau
        6, # v3rhotau2
        3, # v3sigma3
        7, # v3sigma2tau
        8, # v3sigmatau2
        9, # v3tau3
    ]) + 10,
    (0, 4): numpy.array([
        0,  # v4rho4
        1,  # v4rho3sigma
        5,  # v4rho3tau
        2,  # v4rho2sigma2
        6,  # v4rho2sigmatau
        7,  # v4rho2tau2
        3,  # v4rhosigma3
        8,  # v4rhosigma2tau
        9,  # v4rhosigmatau2
        10, # v4rhotau3
        4,  # v4sigma4
        11, # v4sigma3tau
        12, # v4sigma2tau2
        13, # v4sigmatau3
        14, # v4tau4
    ]) + 20,
    (1, 2): numpy.array([
        8, 9, 11, 12, 13, 23, 24, 10, 14, 15, 16, 25, 26, 17, 18, 19, 27, 28,
        20, 21, 29, 30, 22, 31, 32, 33, 34, 35,
    ]),
    (1, 3): numpy.array([
        36, 37, 40, 41, 42, 71, 72, 38, 43, 44, 45, 73, 74, 49, 50, 51, 77, 78,
        52, 53, 79, 80, 54, 81, 82, 89, 90, 91, 39, 46, 47, 48, 75, 76, 55, 56,
        57, 83, 84, 58, 59, 85, 86, 60, 87, 88, 92, 93, 94, 61, 62, 63, 95, 96,
        64, 65, 97, 98, 66, 99, 100, 107, 108, 109, 67, 68, 101, 102, 69, 103,
        104, 110, 111, 112, 70, 105, 106, 113, 114, 115, 116, 117, 118, 119,
    ]),
    (1, 4): numpy.array([
        120, 121, 125, 126, 127, 190, 191, 122, 128, 129, 130, 192, 193, 137,
        138, 139, 198, 199, 140, 141, 200, 201, 142, 202, 203, 216, 217, 218,
        123, 131, 132, 133, 194, 195, 143, 144, 145, 204, 205, 146, 147, 206,
        207, 148, 208, 209, 219, 220, 221, 155, 156, 157, 225, 226, 158, 159,
        227, 228, 160, 229, 230, 249, 250, 251, 161, 162, 231, 232, 163, 233,
        234, 252, 253, 254, 164, 235, 236, 255, 256, 257, 267, 268, 269, 270,
        124, 134, 135, 136, 196, 197, 149, 150, 151, 210, 211, 152, 153, 212,
        213, 154, 214, 215, 222, 223, 224, 165, 166, 167, 237, 238, 168, 169,
        239, 240, 170, 241, 242, 258, 259, 260, 171, 172, 243, 244, 173, 245,
        246, 261, 262, 263, 174, 247, 248, 264, 265, 266, 271, 272, 273, 274,
        175, 176, 177, 275, 276, 178, 179, 277, 278, 180, 279, 280, 295, 296,
        297, 181, 182, 281, 282, 183, 283, 284, 298, 299, 300, 184, 285, 286,
        301, 302, 303, 313, 314, 315, 316, 185, 186, 287, 288, 187, 289, 290,
        304, 305, 306, 188, 291, 292, 307, 308, 309, 317, 318, 319, 320, 189,
        293, 294, 310, 311, 312, 321, 322, 323, 324, 325, 326, 327, 328, 329,
    ])
}

def eval_xc1(xc_code, rho, spin=0, deriv=1, omega=None):
    '''Similar to eval_xc.
    Returns an array with the order of derivatives following xcfun convention.
    '''
    out = _eval_xc(xc_code, rho, spin, deriv=deriv, omega=omega)
    xctype = xc_type(xc_code)
    idx = _libxc_to_xcfun_indices(xctype, spin, deriv)
    return out[idx]

def _libxc_to_xcfun_indices(xctype, spin=0, deriv=1):
    if deriv <= 1:
        return slice(None)
    elif xctype == 'LDA' or xctype == 'HF':
        return slice(None)
    elif xctype == 'GGA':
        if spin == 0:
            return slice(None)
        else:
            idx = [numpy.arange(6)] # up to deriv=1
            for i in range(2, deriv+1):
                idx.append(_GGA_SORT[(spin, i)])
    else: # MGGA
        if spin == 0:
            idx = [numpy.arange(4)] # up to deriv=1
        else:
            idx = [numpy.arange(8)] # up to deriv=1
        for i in range(2, deriv+1):
            idx.append(_MGGA_SORT[(spin, i)])
    return numpy.hstack(idx)

def _eval_xc(xc_code, rho, spin=0, deriv=1, omega=None):
    assert deriv <= max_deriv_order(xc_code)
    xctype = xc_type(xc_code)
    assert xctype in ('HF', 'LDA', 'GGA', 'MGGA')

    rho = numpy.asarray(rho, order='C', dtype=numpy.double)
    if xctype == 'MGGA' and rho.shape[-2] == 6:
        rho = numpy.asarray(rho[...,[0,1,2,3,5],:], order='C')

    hyb, fn_facs = parse_xc(xc_code)
    if omega is not None:
        hyb = hyb[:2] + (float(omega),)

    fn_ids = [x[0] for x in fn_facs]
    facs   = [x[1] for x in fn_facs]
    if hyb[2] != 0:
        # Current implementation does not support different omegas for
        # different RSH functionals if there are multiple RSHs
        omega = [hyb[2]] * len(facs)
    else:
        omega = [0] * len(facs)
    fn_ids_set = set(fn_ids)
    if fn_ids_set.intersection(PROBLEMATIC_XC):
        problem_xc = [PROBLEMATIC_XC[k]
                      for k in fn_ids_set.intersection(PROBLEMATIC_XC)]
        warnings.warn('Libxc functionals %s may have discrepancy to xcfun '
                      'library.\n' % problem_xc)

    if any(needs_laplacian(fid) for fid in fn_ids):
        raise NotImplementedError('laplacian in meta-GGA method')

    nvar, xlen = xc_deriv._XC_NVAR[xctype, spin]
    ngrids = rho.shape[-1]
    rho = rho.reshape(spin+1,nvar,ngrids)
    outlen = lib.comb(xlen+deriv, deriv)
    out = numpy.zeros((outlen,ngrids))
    n = len(fn_ids)
    if n > 0:
        density_threshold = 0
        _itrf.LIBXC_eval_xc(ctypes.c_int(n),
                            (ctypes.c_int*n)(*fn_ids),
                            (ctypes.c_double*n)(*facs),
                            (ctypes.c_double*n)(*omega),
                            ctypes.c_int(spin), ctypes.c_int(deriv),
                            ctypes.c_int(nvar), ctypes.c_int(ngrids),
                            ctypes.c_int(outlen),
                            rho.ctypes.data_as(ctypes.c_void_p),
                            out.ctypes.data_as(ctypes.c_void_p),
                            ctypes.c_double(density_threshold))
    return out

def eval_xc_eff(xc_code, rho, deriv=1, omega=None, spin=None):
    r'''Returns the derivative tensor against the density parameters

    [density_a, (nabla_x)_a, (nabla_y)_a, (nabla_z)_a, tau_a]

    or spin-polarized density parameters

    [[density_a, (nabla_x)_a, (nabla_y)_a, (nabla_z)_a, tau_a],
     [density_b, (nabla_x)_b, (nabla_y)_b, (nabla_z)_b, tau_b]].

    It differs from the eval_xc method in the derivatives of non-local part.
    The eval_xc method returns the XC functional derivatives to sigma
    (|\nabla \rho|^2)

    Args:
        rho: 2-dimensional or 3-dimensional array
            Total density or (spin-up, spin-down) densities (and their
            derivatives if GGA or MGGA functionals) on grids

    Kwargs:
        deriv: int
            derivative orders
        omega: float
            define the exponent in the attenuated Coulomb for RSH functional
        spin : int
            spin polarized if spin > 0
    '''
    xctype = xc_type(xc_code)
    rho = numpy.asarray(rho, order='C', dtype=numpy.double)
    if xctype == 'MGGA' and rho.shape[-2] == 6:
        rho = numpy.asarray(rho[...,[0,1,2,3,5],:], order='C')

    if spin is None:
        spin_polarized = rho.ndim >= 2 and rho.shape[0] == 2
        if spin_polarized:
            spin = 1
        else:
            spin = 0

    out = eval_xc1(xc_code, rho, spin, deriv, omega)
    return xc_deriv.transform_xc(rho, out, xctype, spin, deriv)

def define_xc_(ni, description, xctype='LDA', hyb=0, rsh=(0,0,0)):
    '''Define XC functional.  See also :func:`eval_xc` for the rules of input description.

    Args:
        ni : an instance of :class:`NumInt`

        description : str
            A string to describe the linear combination of different XC functionals.
            The X and C functional are separated by comma like '.8*LDA+.2*B86,VWN'.
            If "HF" was appeared in the string, it stands for the exact exchange.

    Kwargs:
        xctype : str
            'LDA' or 'GGA' or 'MGGA'
        hyb : float
            hybrid functional coefficient
        rsh : a list of three floats
            coefficients (omega, alpha, beta) for range-separated hybrid functional.
            omega is the exponent factor in attenuated Coulomb operator e^{-omega r_{12}}/r_{12}
            alpha is the coefficient for long-range part, hybrid coefficient
            can be obtained by alpha + beta

    Examples:

    >>> mol = gto.M(atom='O 0 0 0; H 0 0 1; H 0 1 0', basis='ccpvdz')
    >>> mf = dft.RKS(mol)
    >>> define_xc_(mf._numint, '.2*HF + .08*LDA + .72*B88, .81*LYP + .19*VWN')
    >>> mf.kernel()
    -76.3783361189611
    >>> define_xc_(mf._numint, 'LDA*.08 + .72*B88 + .2*HF, .81*LYP + .19*VWN')
    >>> mf.kernel()
    -76.3783361189611
    >>> def eval_xc(xc_code, rho, *args, **kwargs):
    ...     exc = 0.01 * rho**2
    ...     vrho = 0.01 * 3 * rho**2
    ...     vxc = (vrho, None, None, None)
    ...     fxc = None  # 2nd order functional derivative
    ...     kxc = None  # 3rd order functional derivative
    ...     return exc, vxc, fxc, kxc
    >>> define_xc_(mf._numint, eval_xc, xctype='LDA')
    >>> mf.kernel()
    48.8525211046668
    '''
    if isinstance(description, str):
        def _eval_xc(xc_code, rho, *args, **kwargs):
            return eval_xc(description, rho, *args, **kwargs)
        ni.eval_xc = _eval_xc
        ni.hybrid_coeff = lambda *args, **kwargs: hybrid_coeff(description)
        ni.rsh_coeff = lambda *args: rsh_coeff(description)
        ni._xc_type = lambda *args: xc_type(description)

    elif callable(description):
        ni.eval_xc = _eval_xc = description
        ni.hybrid_coeff = lambda *args, **kwargs: hyb
        ni.rsh_coeff = lambda *args, **kwargs: rsh
        ni._xc_type = lambda *args: xctype

    else:
        raise ValueError('Unknown description %s' % description)

    def _eval_xc1(xc_code, rho, spin=0, deriv=1, omega=None):
        libxc_out = _eval_xc(xc_code, rho, spin, deriv=deriv, omega=omega)
        nvar, xlen = xc_deriv._XC_NVAR[xctype, spin]
        outlen = lib.comb(xlen+deriv, deriv)
        exc, vxc, fxc, kxc = libxc_out[:4]
        out = [exc]
        if deriv > 0:
            assert vxc is not None
            out.extend([x for x in vxc if x is not None])
        if deriv > 1:
            assert fxc is not None
            if xctype == 'GGA':
                assert len(fxc) == 3, 'fxc for GGA should be arranged as (v2rho2, v2rhosigma, v2sigma2)'
            elif xctype == 'MGGA':
                if len(fxc) == 10:
                    fxc = [fxc[i] for i in [0, 1, 2, 6, 4, 9]]
                else:
                    assert len(fxc) == 6, (
                        'fxc for MGGA should be arranged as\n'
                        '(v2rho2, v2rhosigma, v2sigma2, v2tau2, v2rhotau, v2sigmatau)\nor\n'
                        '(v2rho2, v2rhosigma, v2sigma2, v2lapl2, v2tau2, '
                        'v2rholapl, v2rhotau, v2lapltau, v2sigmalapl, v2sigmatau)')
            assert all(x is not None for x in fxc)
            out.extend(fxc)
        if deriv > 2:
            assert kxc is not None
            out.extend([x for x in kxc if x is not None])

        if spin == 1:
            # Returns of eval_xc are structured as [grid_id,deriv_component]
            # for each term in libxc_out. Change the shape to [deriv_comp, grid_id]
            out = [x.T for x in out]

        out = numpy.vstack(out)[:outlen]
        assert len(out) == outlen
        idx = _libxc_to_xcfun_indices(xctype, spin, deriv)
        return out[idx]
    ni.eval_xc1 = _eval_xc1
    return ni

def define_xc(ni, description, xctype='LDA', hyb=0, rsh=(0,0,0)):
    return define_xc_(ni.copy(), description, xctype, hyb, rsh)
define_xc.__doc__ = define_xc_.__doc__



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Non-relativistic generalized Hartree-Fock with point group symmetry.
'''

from functools import reduce
import numpy
import scipy.linalg
from pyscf import lib
from pyscf import symm
from pyscf.lib import logger
from pyscf.scf import hf
from pyscf.scf import hf_symm
from pyscf.scf import ghf
from pyscf.scf import chkfile
from pyscf import __config__

WITH_META_LOWDIN = getattr(__config__, 'scf_analyze_with_meta_lowdin', True)
MO_BASE = getattr(__config__, 'MO_BASE', 1)


def analyze(mf, verbose=logger.DEBUG, with_meta_lowdin=WITH_META_LOWDIN,
            **kwargs):
    mol = mf.mol
    mo_energy = mf.mo_energy
    mo_occ = mf.mo_occ
    mo_coeff = mf.mo_coeff
    ovlp_ao = mf.get_ovlp()
    log = logger.new_logger(mf, verbose)
    if log.verbose >= logger.NOTE:
        nirrep = len(mol.irrep_id)
        orbsym = mf.get_orbsym(mo_coeff, ovlp_ao)
        wfnsym = 0
        noccs = [sum(orbsym[mo_occ>0]==ir) for ir in mol.irrep_id]
        log.note('total symmetry = %s', symm.irrep_id2name(mol.groupname, wfnsym))
        log.note('occupancy for each irrep:  ' + (' %4s'*nirrep), *mol.irrep_name)
        log.note('double occ                 ' + (' %4d'*nirrep), *noccs)
        log.note('**** MO energy ****')
        irname_full = {}
        for k,ir in enumerate(mol.irrep_id):
            irname_full[ir] = mol.irrep_name[k]
        irorbcnt = {}
        for k, j in enumerate(orbsym):
            if j in irorbcnt:
                irorbcnt[j] += 1
            else:
                irorbcnt[j] = 1
            log.note('MO #%d (%s #%d), energy= %.15g occ= %g',
                     k+MO_BASE, irname_full[j], irorbcnt[j], mo_energy[k],
                     mo_occ[k])

    dm = mf.make_rdm1(mo_coeff, mo_occ)
    dip = mf.dip_moment(mol, dm, verbose=log)
    if with_meta_lowdin:
        pop_and_chg = mf.mulliken_meta(mol, dm, s=ovlp_ao, verbose=log)
    else:
        pop_and_chg = mf.mulliken_pop(mol, dm, s=ovlp_ao, verbose=log)
    return pop_and_chg, dip

def canonicalize(mf, mo_coeff, mo_occ, fock=None):
    '''Canonicalization diagonalizes the UHF Fock matrix in occupied, virtual
    subspaces separatedly (without change occupancy).
    '''
    mol = mf.mol
    if not mol.symmetry:
        raise RuntimeError('mol.symmetry not enabled')

    if getattr(mo_coeff, 'orbsym', None) is not None:
        return hf_symm.canonicalize(mf, mo_coeff, mo_occ, fock)
    else:
        raise NotImplementedError

class SymAdaptedGHF(ghf.GHF):
    __doc__ = ghf.GHF.__doc__ + '''
    Attributes for symmetry allowed GHF:
        irrep_nelec : dict
            Specify the number of electrons for particular irrep
            {'ir_name':int, ...}.
            For the irreps not listed in these dicts, the program will choose the
            occupancy based on the orbital energies.
    '''

    _keys = {'irrep_nelec'}

    def __init__(self, mol):
        ghf.GHF.__init__(self, mol)
        # number of electrons for each irreps
        self.irrep_nelec = {}

    def dump_flags(self, verbose=None):
        ghf.GHF.dump_flags(self, verbose)
        if self.irrep_nelec:
            logger.info(self, 'irrep_nelec %s', self.irrep_nelec)
        return self

    def build(self, mol=None):
        if mol is None: mol = self.mol
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')

        for irname in self.irrep_nelec:
            if irname not in mol.irrep_name:
                logger.warn(self, 'Molecule does not have irrep %s', irname)

        nelec_fix = self.irrep_nelec.values()
        if any(isinstance(x, (tuple, list)) for x in nelec_fix):
            msg =('Number of alpha/beta electrons cannot be assigned '
                  'separately in GHF.  irrep_nelec = %s' % self.irrep_nelec)
            raise ValueError(msg)
        nelec_fix = sum(nelec_fix)
        float_irname = set(mol.irrep_name) - set(self.irrep_nelec)
        if nelec_fix > mol.nelectron:
            msg =('More electrons defined by irrep_nelec than total num electrons. '
                  'mol.nelectron = %d  irrep_nelec = %s' %
                  (mol.nelectron, self.irrep_nelec))
            raise ValueError(msg)
        else:
            logger.info(mol, 'Freeze %d electrons in irreps %s',
                        nelec_fix, self.irrep_nelec.keys())

        if len(float_irname) == 0 and nelec_fix != mol.nelectron:
            msg =('Num electrons defined by irrep_nelec != total num electrons. '
                  'mol.nelectron = %d  irrep_nelec = %s' %
                  (mol.nelectron, self.irrep_nelec))
            raise ValueError(msg)
        else:
            logger.info(mol, '    %d free electrons in irreps %s',
                        mol.nelectron-nelec_fix, ' '.join(float_irname))
        return ghf.GHF.build(self, mol)

    def eig(self, h, s, symm_orb=None, irrep_id=None):
        if symm_orb is None or irrep_id is None:
            mol = self.mol
            symm_orb = mol.symm_orb
            irrep_id = mol.irrep_id

        nirrep = len(symm_orb)
        symm_orb = [scipy.linalg.block_diag(c, c) for c in symm_orb]
        h = symm.symmetrize_matrix(h, symm_orb)
        s = symm.symmetrize_matrix(s, symm_orb)
        cs = []
        es = []
        orbsym = []
        for ir in range(nirrep):
            e, c = self._eigh(h[ir], s[ir])
            cs.append(c)
            es.append(e)
            orbsym.append([irrep_id[ir]] * e.size)
        e = numpy.hstack(es)
        c = hf_symm.so2ao_mo_coeff(symm_orb, cs)
        c = lib.tag_array(c, orbsym=numpy.hstack(orbsym))
        return e, c

    def get_grad(self, mo_coeff, mo_occ, fock=None):
        g = ghf.GHF.get_grad(self, mo_coeff, mo_occ, fock)
        occidx = mo_occ > 0
        viridx = ~occidx
        orbsym = self.get_orbsym(mo_coeff)
        sym_forbid = orbsym[viridx].reshape(-1,1) != orbsym[occidx]
        g[sym_forbid.ravel()] = 0
        return g

    def get_occ(self, mo_energy=None, mo_coeff=None):
        ''' We assumed mo_energy are grouped by symmetry irreps, (see function
        self.eig). The orbitals are sorted after SCF.
        '''
        if mo_energy is None: mo_energy = self.mo_energy
        mol = self.mol
        if not mol.symmetry:
            raise RuntimeError('mol.symmetry not enabled')

        orbsym = self.get_orbsym(mo_coeff)
        mo_occ = numpy.zeros_like(mo_energy)
        rest_idx = numpy.ones(mo_occ.size, dtype=bool)
        nelec_fix = 0
        for i, ir in enumerate(mol.irrep_id):
            irname = mol.irrep_name[i]
            ir_idx = numpy.where(orbsym == ir)[0]
            if irname in self.irrep_nelec:
                n = self.irrep_nelec[irname]
                occ_sort = numpy.argsort(mo_energy[ir_idx].round(9), kind='stable')
                occ_idx  = ir_idx[occ_sort[:n]]
                mo_occ[occ_idx] = 1
                nelec_fix += n
                rest_idx[ir_idx] = False
        nelec_float = mol.nelectron - nelec_fix
        assert (nelec_float >= 0)
        if nelec_float > 0:
            rest_idx = numpy.where(rest_idx)[0]
            occ_sort = numpy.argsort(mo_energy[rest_idx].round(9), kind='stable')
            occ_idx  = rest_idx[occ_sort[:nelec_float]]
            mo_occ[occ_idx] = 1

        vir_idx = (mo_occ==0)
        if self.verbose >= logger.INFO and numpy.count_nonzero(vir_idx) > 0:
            ehomo = max(mo_energy[~vir_idx])
            elumo = min(mo_energy[ vir_idx])
            noccs = []
            for i, ir in enumerate(mol.irrep_id):
                irname = mol.irrep_name[i]
                ir_idx = (orbsym == ir)

                noccs.append(int(mo_occ[ir_idx].sum()))
                if ehomo in mo_energy[ir_idx]:
                    irhomo = irname
                if elumo in mo_energy[ir_idx]:
                    irlumo = irname
            logger.info(self, 'HOMO (%s) = %.15g  LUMO (%s) = %.15g',
                        irhomo, ehomo, irlumo, elumo)

            logger.debug(self, 'irrep_nelec = %s', noccs)
            hf_symm._dump_mo_energy(mol, mo_energy, mo_occ, ehomo, elumo, orbsym,
                                    verbose=self.verbose)

            if mo_coeff is not None and self.verbose >= logger.DEBUG:
                ss, s = self.spin_square(mo_coeff[:,mo_occ>0], self.get_ovlp())
                logger.debug(self, 'multiplicity <S^2> = %.8g  2S+1 = %.8g', ss, s)
        return mo_occ

    def _finalize(self):
        ghf.GHF._finalize(self)

        # Using mergesort because it is stable. We don't want to change the
        # ordering of the symmetry labels when two orbitals are degenerated.
        o_sort = numpy.argsort(self.mo_energy[self.mo_occ> 0].round(9), kind='stable')
        v_sort = numpy.argsort(self.mo_energy[self.mo_occ==0].round(9), kind='stable')
        orbsym = self.get_orbsym(self.mo_coeff)
        self.mo_energy = numpy.hstack((self.mo_energy[self.mo_occ> 0][o_sort],
                                       self.mo_energy[self.mo_occ==0][v_sort]))
        self.mo_coeff = numpy.hstack((self.mo_coeff[:,self.mo_occ> 0].take(o_sort, axis=1),
                                      self.mo_coeff[:,self.mo_occ==0].take(v_sort, axis=1)))
        orbsym = numpy.hstack((orbsym[self.mo_occ> 0][o_sort],
                               orbsym[self.mo_occ==0][v_sort]))
        self.mo_coeff = lib.tag_array(self.mo_coeff, orbsym=orbsym)
        nocc = len(o_sort)
        self.mo_occ[:nocc] = 1
        self.mo_occ[nocc:] = 0
        if self.chkfile:
            chkfile.dump_scf(self.mol, self.chkfile, self.e_tot, self.mo_energy,
                             self.mo_coeff, self.mo_occ, overwrite_mol=False)
        return self

    def analyze(self, verbose=None, **kwargs):
        if verbose is None: verbose = self.verbose
        return analyze(self, verbose, **kwargs)

    @lib.with_doc(hf_symm.get_irrep_nelec.__doc__)
    def get_irrep_nelec(self, mol=None, mo_coeff=None, mo_occ=None, s=None):
        if mol is None: mol = self.mol
        if mo_occ is None: mo_occ = self.mo_occ
        if mo_coeff is None: mo_coeff = self.mo_coeff
        if s is None: s = self.get_ovlp()
        return hf_symm.get_irrep_nelec(mol, mo_coeff, mo_occ, s)

    canonicalize = canonicalize

    def get_orbsym(self, mo_coeff=None, s=None):
        if mo_coeff is None:
            mo_coeff = self.mo_coeff
        if getattr(mo_coeff, 'orbsym', None) is not None:
            return mo_coeff.orbsym
        if s is None:
            s = self.get_ovlp()
        return numpy.asarray(get_orbsym(self.mol, mo_coeff, s))
    orbsym = property(get_orbsym)

    to_gpu = lib.to_gpu

GHF = SymAdaptedGHF


class HF1e(GHF):
    scf = hf._hf1e_scf


def get_orbsym(mol, mo_coeff, s=None, check=False, symm_orb=None, irrep_id=None):
    if symm_orb is None or irrep_id is None:
        symm_orb = mol.symm_orb
        irrep_id = mol.irrep_id
    if mo_coeff is None:
        orbsym = numpy.hstack([[ir] * symm_orb[i].shape[1]
                               for i, ir in enumerate(irrep_id)])
    elif getattr(mo_coeff, 'orbsym', None) is not None:
        orbsym = mo_coeff.orbsym
    else:
        nao = mo_coeff.shape[0] // 2
        if isinstance(s, numpy.ndarray):
            assert (s.size == nao**2 or numpy.allclose(s[:nao,:nao], s[nao:,nao:]))
            s = s[:nao,:nao]
        mo_a = mo_coeff[:nao].copy()
        mo_b = mo_coeff[nao:]
        zero_alpha_idx = numpy.linalg.norm(mo_a, axis=0) < 1e-7
        mo_a[:,zero_alpha_idx] = mo_b[:,zero_alpha_idx]
        orbsym = symm.label_orb_symm(mol, irrep_id, symm_orb,
                                     mo_a, s, check)
    return numpy.asarray(orbsym)


if __name__ == '__main__':
    from pyscf import gto
    mol = gto.Mole()
    mol.build(
        verbose = 1,
        output = None,
        atom = [['O', (0.,0.,0.)],
                ['O', (0.,0.,1.)], ],
        basis = {'O': 'ccpvdz'},
        symmetry = True,
        charge = -1,
        spin = 1
    )

    method = GHF(mol)
    method.verbose = 5
    method.irrep_nelec['A1u'] = 1
    energy = method.kernel()
    print(energy - -126.117033823738)
    method.canonicalize(method.mo_coeff, method.mo_occ)
    method.analyze()



#!/usr/bin/env python
# Copyright 2021-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import copy
import numpy
from scipy.special import erf

from pyscf import lib
from pyscf import gto, scf
from pyscf.dft import gen_grid, numint
from pyscf.pbc import gto as pbcgto
from pyscf.scf import atom_hf, rohf

def get_pp_loc_part1_rs(mol, coords):
    atm_coords = mol.atom_coords()
    out = 0
    for ia in range(mol.natm):
        r0 = atm_coords[ia]
        r2 = numpy.sum((coords - r0)**2, axis=1)
        r = numpy.sqrt(r2)
        Zia = mol.atom_charge(ia)
        symb = mol.atom_symbol(ia)
        if symb in mol._pseudo:
            pp = mol._pseudo[symb]
            rloc, nexp, cexp = pp[1:3+1]
        else:
            rloc = 1e16
        alpha = 1.0 / (numpy.sqrt(2) * rloc)
        out += - Zia / r * erf(alpha * r)
    return out

def _aux_e2(cell, auxcell, intor, aosym='s1', comp=1):
    intor = cell._add_suffix(intor)
    pcell = copy.copy(cell)
    pcell._atm, pcell._bas, pcell._env = \
            atm, bas, env = gto.conc_env(cell._atm, cell._bas, cell._env,
                                         cell._atm, cell._bas, cell._env)
    ao_loc = gto.moleintor.make_loc(bas, intor)
    aux_loc = auxcell.ao_loc_nr(auxcell.cart or 'ssc' in intor)
    ao_loc = numpy.asarray(numpy.hstack([ao_loc, ao_loc[-1]+aux_loc[1:]]),
                           dtype=numpy.int32)
    atm, bas, env = gto.conc_env(atm, bas, env,
                                 auxcell._atm, auxcell._bas, auxcell._env)
    nbas = cell.nbas
    shls_slice = (0, nbas, nbas, nbas*2, nbas*2, nbas*2+auxcell.nbas)
    comp = 1
    out = gto.moleintor.getints3c(intor, atm, bas, env, shls_slice=shls_slice,
                                  comp=comp, aosym=aosym, ao_loc=ao_loc)
    return out

def get_pp_loc_part2(mol):
    buf = 0
    intors = ('int3c2e', 'int3c1e', 'int3c1e_r2_origk',
              'int3c1e_r4_origk', 'int3c1e_r6_origk')
    for cn in range(1, 5):
        fakecell = pbcgto.pseudo.pp_int.fake_cell_vloc(mol, cn)
        if fakecell.nbas > 0:
            v = _aux_e2(mol, fakecell, intors[cn], aosym='s2', comp=1)
            buf += numpy.einsum('...i->...', v)
    if numpy.isscalar(buf):
        vpp_loc =  buf
    else:
        vpp_loc = lib.unpack_tril(buf)
    return vpp_loc

def get_pp_loc(mol):
    # TODO use analytic integral
    grids = gen_grid.Grids(mol)
    grids.level = 3
    grids.build(with_non0tab=True)
    _numint = numint.NumInt()

    vpp = 0
    for ao, mask, weight, coords in _numint.block_loop(mol, grids):
        vloc = get_pp_loc_part1_rs(mol, coords)
        vpp += numpy.einsum("g,g,gi,gj->ij", weight, vloc, ao, ao)
    vpp += get_pp_loc_part2(mol)
    return vpp

def get_pp_nl(mol):
    nao = mol.nao
    fakecell, hl_blocks = pbcgto.pseudo.pp_int.fake_cell_vnl(mol)
    ppnl_half = _int_vnl(mol, fakecell, hl_blocks)

    ppnl = numpy.zeros((nao,nao), dtype=numpy.double)
    offset = [0] * 3
    for ib, hl in enumerate(hl_blocks):
        l = fakecell.bas_angular(ib)
        nd = 2 * l + 1
        hl_dim = hl.shape[0]
        ilp = numpy.ndarray((hl_dim,nd,nao), dtype=numpy.double)
        for i in range(hl_dim):
            p0 = offset[i]
            ilp[i] = ppnl_half[i][p0:p0+nd]
            offset[i] = p0 + nd
        ppnl += numpy.einsum('ilp,ij,jlq->pq', ilp, hl, ilp)
    return ppnl

def _int_vnl(cell, fakecell, hl_blocks):
    intopt = lib.c_null_ptr()

    def int_ket(_bas, intor):
        if len(_bas) == 0:
            return []
        intor = cell._add_suffix(intor)
        atm, bas, env = gto.conc_env(cell._atm, cell._bas, cell._env,
                                     fakecell._atm, _bas, fakecell._env)
        atm = numpy.asarray(atm, dtype=numpy.int32)
        bas = numpy.asarray(bas, dtype=numpy.int32)
        env = numpy.asarray(env, dtype=numpy.double)
        nbas = len(bas)
        shls_slice = (cell.nbas, nbas, 0, cell.nbas)
        ao_loc = gto.moleintor.make_loc(bas, intor)
        ni = ao_loc[shls_slice[1]] - ao_loc[shls_slice[0]]
        nj = ao_loc[shls_slice[3]] - ao_loc[shls_slice[2]]
        out = numpy.empty((ni,nj), dtype=numpy.double)
        comp = 1
        out = gto.moleintor.getints2c(intor, atm, bas, env, shls_slice=shls_slice, comp=comp, hermi=0,
                                      ao_loc=ao_loc, cintopt=intopt, out=out)
        return out

    hl_dims = numpy.asarray([len(hl) for hl in hl_blocks])
    out = (int_ket(fakecell._bas[hl_dims>0], 'int1e_ovlp'),
           int_ket(fakecell._bas[hl_dims>1], 'int1e_r2_origi'),
           int_ket(fakecell._bas[hl_dims>2], 'int1e_r4_origi'))
    return out

class AtomSCFPP(atom_hf.AtomSphAverageRHF):
    def get_hcore(self, mol=None):
        if mol is None:
            mol = self.mol
        h = mol.intor('int1e_kin', hermi=1)
        h += get_pp_nl(mol)
        h += get_pp_loc(mol)
        return h

class AtomHF1ePP(rohf.HF1e, AtomSCFPP):
    eig = AtomSCFPP.eig
    get_hcore = AtomSCFPP.get_hcore



# Author: Daniel S. King

'''
APC Ranked-Orbital Active Space Selection
If you find this module useful for your work, please consider citing the following:

A Ranked-Orbital Approach to Select Active Spaces for High-Throughput Multireference Computation
https://doi.org/10.1021/acs.jctc.1c00037

Large-Scale Benchmarking of Multireference Vertical-Excitation Calculations via Automated Active-Space Selection
https://doi.org/10.1021/acs.jctc.2c00630
'''

from pyscf.lib import logger
from pyscf import scf, lib
import numpy as np

class Chooser():
    """
    Chooser Class
    Implements the ranked-orbital selection scheme outlined in https://doi.org/10.1021/acs.jctc.1c00037
    Given a set of entropies, will select all orbitals for the active space and then drop the lowest-entropy orbitals
    until the size constraint max_size is met.

    Args:
        orbs: 2D Numpy Array
            Orbitals to choose from, spanning the entire basis (must be square matrix of coefficients)
        occ: 1D Numpy Array
            Orbital occupations for orbs (2,1,0); nactel will be set to the number of electrons in the selected orbitals
        entropies: 1D Numpy Array
            Importance measurement used to rank the orbitals
        max_size: Int or Tuple
            Active space size constraint.
            If tuple, interpreted as (nelecas,ncas)
            If int, interpreted as max # of orbitals

    Returns:
        active-space-size, #-active-electrons, orbital-initial-guess, chosen-active-orbital-indices

    Example:

    #Randomly ranked orbitals
    >>> import numpy as np
    >>> from pyscf import gto, scf, mcscf
    >>> from pyscf.mcscf import apc
    >>> mol = gto.M(atom='H 0 0 0; H 0 0 1', basis='ccpvtz')
    >>> mf = scf.RHF(mol).run()
    >>> entropies = np.random.choice(np.arange(len(mf.mo_occ)),len(mf.mo_occ),replace=False)
    >>> chooser = apc.Chooser(mf.mo_coeff,mf.mo_occ,entropies,max_size=(2,2))
    >>> ncas, nelecas, casorbs, active_idx = chooser.kernel()
    >>> mc = mcscf.CASSCF(mf, ncas, nelecas).run(casorbs)
    """

    def __init__(self,orbs,occ,entropies,max_size=(8,8),verbose=4):
        #Check that we have a full set of orbitals:
        assert(orbs.shape[0] == orbs.shape[1])
        assert(len(occ) == len(entropies))
        assert(len(entropies) == orbs.shape[1])

        self.log = logger.new_logger(lib.StreamObject,verbose)
        self.orbs = orbs
        self.occ = np.array(occ)
        self.entropies = np.asarray(entropies)
        self.max_size = max_size
        self.verbose = verbose

    def _ncsf(self,nactel,norbs):
        """
        Returns number of CSFs in a (nactel,nactorbs) active space
        Assumes minimum number Sz = alpha - beta (0 for even nactel, 1 for odd nactel)
        """
        from scipy.special import comb
        alpha = int(nactel//2 + nactel%2)
        beta = int(nactel//2)
        term1 = comb(norbs,alpha)*comb(norbs,beta)
        term2 = comb(norbs,alpha+1)*comb(norbs,beta-1)
        ncsf = term1-term2
        return ncsf

    def _calc_ncsf(self,active_idx):
        """
        Returns the number of CSFs given the active index using the info in self.occ
        Passes this info to self._ncsf to calculate the size of the active space
        """
        occ = self.occ
        nactel = np.sum(np.array(occ)[active_idx])
        norbs = len(active_idx)
        return self._ncsf(nactel,norbs)

    def _as_is_reasonable(self,active_idx):
        #Checks active space reasonability
        occ = self.occ

        nactel = np.sum(np.array(occ)[active_idx])
        num_os = len(np.where(occ == 1)[0])
        nactorbs = len(active_idx)

        condition1 = (nactel > 0)
        condition2 = (nactel < 2*len(active_idx))
        condition3 = (nactorbs >= num_os)

        if (condition1 and condition2 and condition3):
            return True
        else:
            self.log.debug("Active space is not reasonable!")
            self.log.debug(f"Nactel: {nactel}, Nactorbs: {nactorbs}, Num OS: {num_os}")
            if not condition1:
                self.log.debug("Condition 1 not met")
            elif not condition2:
                self.log.debug("Condition 2 not met")
            elif not condition3:
                self.log.debug("Condition 3 not met")
            return False

    def kernel(self):
        log = self.log
        entropies = self.entropies.copy()
        occ = self.occ.copy()

        #Change singly occupied orbitals to have larger entropies so they are selected:
        os_idx = np.where(occ == 1)[0]
        entropies[os_idx] = np.max(entropies) + 0.01
        if len(os_idx) > 0:
            log.info("Singly occupied orbitals found, setting them to have entropy of max + 0.01...")

        #Start with all orbitals in active space:
        active_idx = list(range(len(entropies)))
        inactive_idx = []
        secondary_idx = []

        #Size constraint:
        if isinstance(self.max_size, (tuple, list, np.ndarray)):
            nactel,norbs = self.max_size
            max_size = self._ncsf(nactel,norbs)
            as_size = self._calc_ncsf(active_idx)
        else:
            max_size = self.max_size
            as_size = len(active_idx)

        nactel = int(np.sum(np.array(occ)[active_idx]))
        nactorbs = len(active_idx)
        log.debug(f"Initial active space of ({nactel},{nactorbs}) has size {as_size}")
        log.debug(f"Maximum active space size set to {max_size}")

        #Drop orbitals until size constraint is satisfied:
        while as_size > max_size:
            nactel = int(np.sum(np.array(occ)[active_idx]))
            nactorbs = len(active_idx)
            log.debug(f"Active space of ({nactel},{nactorbs}) has size {as_size} larger than {max_size}")
            log.debug("Dropping lowest entropy orbital...")

            #Get active orbital entropies
            active_entropies = entropies[active_idx]
            ranked_active_idx = [active_idx[i] for i in np.argsort(active_entropies)]

            #Drop lowest orbital in succession, checking for reasonability:
            active_space_is_reasonable = False
            tries = 0

            while not active_space_is_reasonable:
                try:
                    dropped_idx = ranked_active_idx[tries] #Move to next possibility
                    dropped_idx_entropy = np.round(entropies[dropped_idx],4)
                    dropped_idx_occ = int(occ[dropped_idx])
                except IndexError:
                    log.error("Not enough orbitals to choose a reasonable active space!")
                    raise RuntimeError("Not enough orbitals to choose a reasonable active space!")

                new_inactive_idx = inactive_idx.copy()
                new_active_idx = active_idx.copy()
                new_secondary_idx = secondary_idx.copy()

                if dropped_idx_occ > 0:
                    new_active_idx.remove(dropped_idx)
                    new_inactive_idx += [dropped_idx]
                else:
                    new_active_idx.remove(dropped_idx)
                    new_secondary_idx += [dropped_idx]

                log.debug(f"Attempting to drop orbital {dropped_idx} \
                (occ={dropped_idx_occ}, S={dropped_idx_entropy})...")
                active_space_is_reasonable = self._as_is_reasonable(new_active_idx)
                if active_space_is_reasonable:
                    log.debug("Orbital has been dropped")
                else:
                    log.debug("Active space becomes unreasonable if this orbital is dropped, trying next option...")

                tries += 1

            inactive_idx = new_inactive_idx
            active_idx = new_active_idx
            secondary_idx = new_secondary_idx

            #Calculate new NCSFs:
            if isinstance(self.max_size,tuple):
                nactel,norbs = self.max_size
                as_size = self._calc_ncsf(active_idx)
            else:
                as_size = len(active_idx)

        #Final checks:
        assert(len(active_idx) <= len(entropies))
        assert(as_size <= max_size)

        orbs = self.orbs.copy()
        inactive_orbs = orbs[:,inactive_idx]
        active_orbs = orbs[:,active_idx]
        secondary_orbs = orbs[:,secondary_idx]
        casorbs = np.hstack([inactive_orbs,active_orbs,secondary_orbs])

        nactorbs = active_orbs.shape[1]
        active_occ = np.array(occ)[active_idx]
        nboth = int(np.sum(active_occ[np.where(active_occ == 2)])/2)
        nalpha = int(np.sum(active_occ[np.where(active_occ == 1)]))
        alpha = nboth + nalpha
        beta = nboth
        nactel = (alpha,beta)

        log.info(f"Final selected active space: ({nactel},{nactorbs})")

        return nactorbs, nactel, casorbs, active_idx

class APC():

    """
    APC Class
    Implements APC orbital entropy estimation from https://doi.org/10.1021/acs.jctc.1c00037
    APC-N implemented from https://doi.org/10.1021/acs.jctc.2c00630

    .kernel() combines this with the ranked-orbital scheme implemented in Chooser() to select
    an active space of size max_size from the orbitals in mf.mo_coeff with occupancy mf.mo_occ

    Args:
        mf: an :class:`SCF` object
            Must expose mf.mo_coeff, mf.mo_occ, mf.get_fock(), and mf.get_k()
        max_size: Int or Tuple
            Active space size constraint.
            If tuple, interpreted as (nelecas,ncas)
            If int interpreted as max # of orbitals
        n: Int
            Number of times to remove highest-entropy virtual orbitals in entropy calculation.
            A higher value will tend to select active spaces with less doubly occupied orbitals.

    Kwargs:
        eps: Float
            Small offset added to singly occupied and removed virtual orbital entropies (can generally be ignored)

    Returns:
        active-space-size, #-active-electrons, orbital-initial-guess (following AVAS convention)

    Example:
    >>> import numpy as np
    >>> from pyscf import gto, scf, mcscf
    >>> from pyscf.mcscf import apc
    >>> mol = gto.M(atom='H 0 0 0; H 0 0 1', basis='ccpvtz')
    >>> mf = scf.RHF(mol).run()
    >>> myapc = apc.APC(mf,max_size=2)
    >>> ncas,nelecas,casorbs = myapc.kernel()
    >>> mc = mcscf.CASSCF(mf, ncas, nelecas).run(casorbs)
    """

    def __init__(self,mf,max_size=(8,8),n=2,eps=1e-3,verbose=4):
        self.log = logger.new_logger(lib.StreamObject,verbose)
        self.mf = mf
        self.n = n
        self.eps = eps
        assert(eps > 0) #Check that eps > 0
        self.max_size = max_size
        self.verbose = verbose

    def _apc(self,orbs,occ,f_mo,k_mo):
        """
        Calculates APC entropies for given orbitals, occupations, and F and K matrix elements
        Singly occupied orbitals are set to max value of other orbitals + self.eps

        Args:
            orbs: 2D Numpy Array
                A nbasis x nmo array of candidate AS orbitals
            occ: 1D Numpy Array
                Orbital occupations for orbs (2,1,0)
            f_mo: 2D Numpy Array
                Fock operator in the basis of the orbs (nmo x nmo)
            k_mo: 2D Numpy Array
                Exchange operator in the basis of the orbs (nmo x nmo)
        """
        eps = self.eps
        docc_idx = np.where(occ == 2)[0]
        os_idx = np.where(occ == 1)[0]
        virt_idx = np.where(occ == 0)[0]

        #Calculate APCs
        apcs = np.zeros([len(docc_idx),len(virt_idx)])
        for i,d in enumerate(docc_idx):
            for j,v in enumerate(virt_idx):
                k12 = 0.5*k_mo[v,v]
                delta = f_mo[v,v] - f_mo[d,d]
                c = -k12/(delta + np.sqrt(k12**2 + delta**2))
                apcs[i,j] = c

        #Calculate entropies
        apc_entropies = np.zeros(orbs.shape[1])
        for o in range(orbs.shape[1]):

            #Collect APCs for this orbital:
            if o in os_idx:
                continue #Fill in later with max value
            elif o in docc_idx:
                idx = np.where(docc_idx == o)[0][0]
                apcs_o = apcs[idx,:]
            elif o in virt_idx:
                idx = np.where(virt_idx == o)[0][0]
                apcs_o = apcs[:,idx]

            #Normalize APCs:
            cis = apcs_o
            cis2 = cis**2
            sumci2 = np.sum(cis2)
            norm = np.sqrt((sumci2 + 1))
            cisnorm = cis/norm

            #Square Normalized APCs to calculate entropies:
            cisnorm2 = cisnorm**2
            assert((cisnorm2 < 1).any().all())
            sumcisnorm2 = np.sum(cisnorm2)
            assert(np.allclose((sumcisnorm2 + (1/norm)**2),1,atol=1e-6))
            exent = -sumcisnorm2 * np.log(sumcisnorm2)
            gsent = -(1/norm)**2 * np.log((1/norm)**2)
            ent = exent + gsent
            apc_entropies[o] = ent

        #Assign max value to singly occupied orbitals plus some small value:
        apc_entropies[os_idx] = np.max(apc_entropies) + eps

        return apc_entropies

    def _calc_apc_entropies(self,mf):
        """
        Implements the "APC-N" approach in which high-entropy virtual orbitals are repeatedly set to singly occupied
        Then sets the singly occupied orbitals and previously removed orbitals to high values
        Reads the value of n from self.n

        Args:
            mf: an :class:`SCF` object
                Must expose mf.mo_coeff, mf.mo_occ, mf.get_fock(), and mf.get_k()
        """

        log = self.log
        n = self.n
        eps = self.eps
        log.info(f"Calculating APC entropies (N={n})...")

        f_ao = mf.get_fock()
        k_ao = mf.get_k()

        if isinstance(mf, scf.uhf.UHF):
            log.note('UHF object found. APC uses averaged F, summed K, summed occupation, and alpha orbitals.')
            orbs = mf.mo_coeff[0]
            occ = mf.mo_occ.sum(axis=0) #summed occupation
            f_ao = np.sum(f_ao,axis=0)/2 #averaged fock
            k_ao = np.sum(k_ao,axis=0) #summed exchange
        elif isinstance(mf, scf.rohf.ROHF):
            log.note('ROHF object found. APC uses summed K')
            orbs = mf.mo_coeff
            occ = mf.mo_occ.copy()
            k_ao = np.sum(k_ao,axis=0) #summed exchange
        else:
            orbs = mf.mo_coeff
            occ = mf.mo_occ.copy()

        #Calculate f and k in mo basis
        log.info("Transforming F and K to MO basis...")
        f_mo = np.linalg.multi_dot([orbs.T,f_ao,orbs])
        k_mo = np.linalg.multi_dot([orbs.T,k_ao,orbs])

        #Calculate entropies
        removed_idx = []
        original_os = np.where(occ == 1)[0]

        log.info("Calculating initial APC entropies...")
        apc_entropies = self._apc(orbs,occ,f_mo,k_mo)

        for loop_n in range(n):
            if loop_n > 0:
                log.info(f"Calculating APC entropies (Round {loop_n})...")
                apc_entropies = self._apc(orbs,occ,f_mo,k_mo)
            maxS = np.round(np.max(apc_entropies),5)
            log.info(f"Maximum entropy: {maxS}")

            #Remove highest virtual and set occ to 1
            virt_idx = np.where(occ == 0)[0]
            to_remove = virt_idx[np.argmax(apc_entropies[virt_idx])]
            removed_idx += [to_remove]
            log.info(f"Setting maximum virtual orbitals {removed_idx} to occupation 1...")
            occ[removed_idx] = 1

        log.info("Calculating final APC entropies...")
        apc_entropies = self._apc(orbs,occ,f_mo,k_mo)
        maxS = np.round(np.max(apc_entropies),5)
        log.info(f"Maximum entropy: {maxS}")

        #Iterate over os and removed virtuals and set to max in order:
        maxs = np.max(apc_entropies)
        for i,o in enumerate(original_os):
            apc_entropies[o] = maxs + 2*eps - i*eps*1e-2
        for i,o in enumerate(removed_idx):
            apc_entropies[o] = maxs + eps - i*eps*1e-2

        return apc_entropies

    def kernel(self):
        log = self.log
        log.info('\n** APC Active Space Selection **')
        entropies = self._calc_apc_entropies(self.mf)
        self.entropies = entropies

        if isinstance(self.mf, scf.uhf.UHF):
            orbs = self.mf.mo_coeff[0] #alpha orbitals
            occ = self.mf.mo_occ.sum(axis=0) #summed occupation
        else:
            orbs = self.mf.mo_coeff
            occ = self.mf.mo_occ

        max_size = self.max_size
        log.info(f"Choosing active space with ranked orbital approach (max_size = {max_size})...")
        chooser = Chooser(orbs,occ,entropies,max_size,verbose=self.verbose)
        nactorbs, nactel, casorbs, active_idx = chooser.kernel()
        self.active_idx = active_idx
        return nactorbs, nactel, casorbs



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Timothy Berkelbach <tim.berkelbach@gmail.com>
#          Qiming Sun <osirpt.sun@gmail.com>
#

'''
Non-relativistic Restricted Kohn-Sham for periodic systems at a single k-point

See Also:
    pyscf.pbc.dft.krks.py : Non-relativistic Restricted Kohn-Sham for periodic
                            systems with k-point sampling
'''


import numpy
import pyscf.dft
from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.scf import hf as pbchf
from pyscf.pbc.scf import khf
from pyscf.pbc.scf import addons
from pyscf.pbc.dft import gen_grid
from pyscf.pbc.dft import numint
from pyscf.dft import rks as mol_ks
from pyscf.pbc.dft import multigrid
from pyscf.pbc.lib.kpts import KPoints
from pyscf import __config__


def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpt=None, kpts_band=None):
    '''Coulomb + XC functional

    .. note::
        This function will change the ks object.

    Args:
        ks : an instance of :class:`RKS`
            XC functional are controlled by ks.xc attribute.  Attribute
            ks.grids might be initialized.
        dm : ndarray or list of ndarrays
            A density matrix or a list of density matrices

    Returns:
        matrix Veff = J + Vxc.  Veff can be a list matrices, if the input
        dm is a list of density matrices.
    '''
    if cell is None: cell = ks.cell
    if dm is None: dm = ks.make_rdm1()
    if kpt is None: kpt = ks.kpt
    t0 = (logger.process_clock(), logger.perf_counter())

    ni = ks._numint
    if isinstance(ni, multigrid.MultiGridNumInt):
        if ks.do_nlc():
            raise NotImplementedError(f'MultiGrid for NLC functional {ks.xc} + {ks.nlc}')
        j_in_xc = ni.xc_with_j
    else:
        ks.initialize_grids(cell, dm, kpt)
        j_in_xc = False

    max_memory = ks.max_memory - lib.current_memory()[0]
    n, exc, vxc = ni.nr_rks(cell, ks.grids, ks.xc, dm, 0, hermi,
                            kpt, kpts_band, max_memory=max_memory)
    logger.info(ks, 'nelec by numeric integration = %s', n)
    if ks.do_nlc():
        if ni.libxc.is_nlc(ks.xc):
            xc = ks.xc
        else:
            assert ni.libxc.is_nlc(ks.nlc)
            xc = ks.nlc
        n, enlc, vnlc = ni.nr_nlc_vxc(cell, ks.nlcgrids, xc, dm, 0, hermi, kpt,
                                      max_memory=max_memory)
        exc += enlc
        vxc += vnlc
        logger.info(ks, 'nelec with nlc grids = %s', n)
    t0 = logger.timer(ks, 'vxc', *t0)

    ground_state = kpts_band is None
    vj, vk = _get_jk(ks, cell, dm, hermi, kpt, kpts_band, with_j=not j_in_xc)
    if j_in_xc:
        ecoul = vxc.ecoul
    else:
        vxc += vj
        ecoul = None
        if ground_state:
            ecoul = numpy.einsum('ij,ji', dm, vj).real * .5
    if ni.libxc.is_hybrid_xc(ks.xc):
        vxc -= .5 * vk
        if ground_state:
            exc -= numpy.einsum('ij,ji->', dm, vk).real * .25
    vxc = lib.tag_array(vxc, ecoul=ecoul, exc=exc, vj=None, vk=None)
    logger.timer(ks, 'veff', *t0)
    return vxc

def _patch_df_beckegrids(density_fit):
    def new_df(self, auxbasis=None, with_df=None, *args, **kwargs):
        mf = density_fit(self, auxbasis, with_df, *args, **kwargs)
        mf.with_df._j_only = not self._numint.libxc.is_hybrid_xc(self.xc)
        mf.grids = gen_grid.BeckeGrids(self.cell)
        mf.grids.level = getattr(__config__, 'dft_rks_RKS_grids_level',
                                 mf.grids.level)
        mf.nlcgrids = gen_grid.BeckeGrids(self.cell)
        mf.nlcgrids.level = getattr(__config__, 'dft_rks_RKS_nlcgrids_level',
                                    mf.nlcgrids.level)
        return mf
    return new_df

NELEC_ERROR_TOL = getattr(__config__, 'pbc_dft_rks_prune_error_tol', 0.02)
def prune_small_rho_grids_(ks, cell, dm, grids, kpts):
    rho = ks.get_rho(dm, grids, kpts)
    n = numpy.dot(rho, grids.weights)
    if abs(n-cell.nelectron) < NELEC_ERROR_TOL*n:
        rho *= grids.weights
        idx = abs(rho) > ks.small_rho_cutoff / grids.weights.size
        logger.debug(ks, 'Drop grids %d',
                     grids.weights.size - numpy.count_nonzero(idx))
        grids.coords  = numpy.asarray(grids.coords [idx], order='C')
        grids.weights = numpy.asarray(grids.weights[idx], order='C')
        grids.non0tab = grids.make_mask(cell, grids.coords)
    return grids

@lib.with_doc(pbchf.get_rho.__doc__)
def get_rho(mf, dm=None, grids=None, kpt=None):
    if dm is None: dm = mf.make_rdm1()
    if grids is None: grids = mf.grids
    if kpt is None: kpt = mf.kpt
    if dm[0].ndim == 2:  # the UKS density matrix
        dm = dm[0] + dm[1]
    return mf._numint.get_rho(mf.cell, dm, grids, kpt, mf.max_memory)

def gen_response(mf, mo_coeff=None, mo_occ=None,
                 singlet=None, hermi=0, max_memory=None, with_nlc=True):
    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ
    cell = mf.cell
    kpt = mf.kpt
    ni = mf._numint
    hybrid = ni.libxc.is_hybrid_xc(mf.xc)
    j_in_xc = getattr(ni, 'xc_with_j', False)

    if with_nlc and mf.do_nlc():
        raise NotImplementedError

    if singlet is None:  # for newton solver
        spin = 0
    else:
        spin = 1
    rho0, vxc, fxc = ni.cache_xc_kernel(cell, mf.grids, mf.xc, mo_coeff,
                                        mo_occ, spin, kpt)
    dm0 = None

    if max_memory is None:
        mem_now = lib.current_memory()[0]
        max_memory = max(2000, mf.max_memory*.8-mem_now)

    if singlet is None:  # Without specify singlet, general case
        def vind(dm1):
            # The singlet hessian
            if hermi == 2:
                v1 = numpy.zeros_like(dm1)
            else:
                v1 = ni.nr_rks_fxc(cell, mf.grids, mf.xc, dm0, dm1, 0, hermi,
                                   rho0, vxc, fxc, kpt, max_memory=max_memory)
            vj, vk = _get_jk(mf, cell, dm1, hermi, kpt, with_j=not j_in_xc)
            if not j_in_xc:
                v1 += vj
            if hybrid:
                v1 -= .5 * vk
            return v1

    elif singlet:
        fxc *= .5
        def vind(dm1):
            if hermi == 2:
                v1 = numpy.zeros_like(dm1)
            else:
                # nr_rks_fxc_st requires alpha of dm1
                v1 = ni.nr_rks_fxc_st(cell, mf.grids, mf.xc, dm0, dm1, hermi,
                                      True, rho0, vxc, fxc, kpt,
                                      max_memory=max_memory)
            vj, vk = _get_jk(mf, cell, dm1, hermi, kpt, with_j=not j_in_xc)
            if not j_in_xc:
                v1 += vj
            if hybrid:
                v1 -= .5 * vk
            return v1
    else:  # triplet
        fxc *= .5
        def vind(dm1):
            if hermi == 2:
                v1 = numpy.zeros_like(dm1)
            else:
                # nr_rks_fxc_st requires alpha of dm1
                v1 = ni.nr_rks_fxc_st(cell, mf.grids, mf.xc, dm0, dm1, hermi,
                                      False, rho0, vxc, fxc, kpt,
                                      max_memory=max_memory)
            vk = _get_jk(mf, cell, dm1, hermi, kpt, with_j=False)[1]
            if hybrid:
                v1 -= .5 * vk
            return v1
    return vind

def _get_jk(mf, cell, dm, hermi, kpt, kpts_band=None, with_j=True):
    '''J and Exx matrix. Note, Exx here is a scaled HF K term.'''
    ni = mf._numint
    omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, spin=cell.spin)
    hybrid = ni.libxc.is_hybrid_xc(mf.xc)
    vj = vk = None
    if not hybrid:
        if hermi == 2 or not with_j:
            vj = 0
        else:
            vj = mf.get_j(cell, dm, hermi, kpt, kpts_band)
        return vj, vk

    if omega == 0:
        if hermi == 2 or not with_j:
            vk = mf.get_k(cell, dm, hermi, kpt, kpts_band)
        else:
            vj, vk = mf.get_jk(cell, dm, hermi, kpt, kpts_band)
        vk *= hyb
    elif alpha == 0: # LR=0, only SR exchange
        vk = mf.get_k(cell, dm, hermi, kpt, kpts_band, omega=-omega)
        vk *= hyb
        vj = mf.get_j(cell, dm, hermi, kpt, kpts_band)
    elif hyb == 0: # SR=0, only LR exchange
        vk = mf.get_k(cell, dm, hermi, kpt, kpts_band, omega=omega)
        vk *= alpha
        vj = mf.get_j(cell, dm, hermi, kpt, kpts_band)
    else: # SR and LR exchange with different ratios
        if hermi == 2 or not with_j:
            vk = mf.get_k(cell, dm, hermi, kpt, kpts_band)
        else:
            vj, vk = mf.get_jk(cell, dm, hermi, kpt, kpts_band)
        vk *= hyb
        vklr = mf.get_k(cell, dm, hermi, kpt, kpts_band, omega=omega)
        vklr *= (alpha - hyb)
        vk += vklr
    return vj, vk

class KohnShamDFT(mol_ks.KohnShamDFT):
    '''PBC-KS'''

    _keys = {'xc', 'nlc', 'grids', 'nlcgrids', 'small_rho_cutoff'}
    # Use rho to filter grids
    small_rho_cutoff = getattr(__config__, 'dft_rks_RKS_small_rho_cutoff', 1e-7)

    get_rho = get_rho

    density_fit = _patch_df_beckegrids(pbchf.RHF.density_fit)
    rs_density_fit = _patch_df_beckegrids(pbchf.RHF.rs_density_fit)
    mix_density_fit = _patch_df_beckegrids(pbchf.RHF.mix_density_fit)
    gen_response = NotImplemented

    def __init__(self, xc='LDA,VWN'):
        self.xc = xc
        self.grids = gen_grid.UniformGrids(self.cell)
        self.nlc = ''
        self.nlcgrids = gen_grid.UniformGrids(self.cell)
##################################################
# don't modify the following attributes, they are not input options
        # Note Do not refer to .with_df._numint because mesh/coords may be different
        if isinstance(self, khf.KSCF):
            if isinstance(self.kpts, KPoints):
                self._numint = numint.KNumInt(self.kpts.kpts)
            else:
                self._numint = numint.KNumInt(self.kpts)
        else:
            self._numint = numint.NumInt()

    def dump_flags(self, verbose=None):
        logger.info(self, 'XC functionals = %s', self.xc)
        logger.info(self, 'small_rho_cutoff = %g', self.small_rho_cutoff)
        self.grids.dump_flags(verbose)
        return self

    def reset(self, cell=None):
        pbchf.SCF.reset(self, cell)
        self.grids.reset(cell)
        self.nlcgrids.reset(cell)
        return self

    def build(self, cell=None):
        # To handle the attribute kpt or kpts loaded from chkfile
        if 'kpts' in self.__dict__:
            self.kpts = self.__dict__.pop('kpts')
        elif 'kpt' in self.__dict__:
            self.kpt = self.__dict__.pop('kpt')

        kpts = self.kpts
        if self.rsjk:
            if not numpy.all(self.rsjk.kpts == self.kpt):
                self.rsjk = self.rsjk.__class__(cell, kpts)

        if self.verbose >= logger.WARN:
            self.check_sanity()
        return self

    @lib.with_doc(pbchf.SCF.jk_method.__doc__)
    def jk_method(self, J='FFTDF', K=None):
        if K is None:
            K = J
        if (('RS' in J or 'RS' in K) and
            not isinstance(self.grids, gen_grid.BeckeGrids)):
            self.grids = gen_grid.BeckeGrids(self.cell)
        return pbchf.SCF.jk_method(self, J, K)

    def to_rks(self, xc=None):
        '''Convert the input mean-field object to a RKS/ROKS object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        mf = addons.convert_to_rhf(self)
        if xc is not None:
            mf.xc = xc
        mf.converged = xc == self.xc and isinstance(self, RKS)
        return mf

    def to_uks(self, xc=None):
        '''Convert the input mean-field object to a UKS object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        mf = addons.convert_to_uhf(self)
        if xc is not None:
            mf.xc = xc
        mf.converged = xc == self.xc
        return mf

    def to_gks(self, xc=None):
        '''Convert the input mean-field object to a GKS object.

        Note this conversion only changes the class of the mean-field object.
        The total energy and wave-function are the same as them in the input
        mean-field object.
        '''
        from pyscf.pbc.dft.numint2c import NumInt2C, KNumInt2C
        mf = addons.convert_to_ghf(self)
        if xc is not None:
            mf.xc = xc
        mf.converged = xc == self.xc
        if isinstance(mf, khf.KSCF):
            if not isinstance(mf._numint, KNumInt2C):
                mf._numint = KNumInt2C(mf._numint.kpts)
        else:
            if not isinstance(mf._numint, NumInt2C):
                mf._numint = NumInt2C()
        return mf

    def to_hf(self):
        raise NotImplementedError

    def initialize_grids(self, cell, dm, kpts, ground_state=True):
        '''Initialize self.grids the first time call get_veff'''
        if self.grids.coords is None:
            t0 = (logger.process_clock(), logger.perf_counter())
            self.grids.build(with_non0tab=True)
            if (isinstance(self.grids, gen_grid.BeckeGrids) and
                self.small_rho_cutoff > 1e-20 and ground_state):
                self.grids = prune_small_rho_grids_(
                    self, self.cell, dm, self.grids, kpts)
            t0 = logger.timer(self, 'setting up grids', *t0)
        is_nlc = self.do_nlc()
        if is_nlc and self.nlcgrids.coords is None:
            t0 = (logger.process_clock(), logger.perf_counter())
            self.nlcgrids.build(with_non0tab=True)
            if (isinstance(self.grids, gen_grid.BeckeGrids) and
                self.small_rho_cutoff > 1e-20 and ground_state):
                self.nlcgrids = prune_small_rho_grids_(
                    self, self.cell, dm, self.nlcgrids, kpts)
            t0 = logger.timer(self, 'setting up nlc grids', *t0)
        return self

    def to_gpu(self):
        raise NotImplementedError

# Update the KohnShamDFT label in pbc.scf.hf module
pbchf.KohnShamDFT = KohnShamDFT


class RKS(KohnShamDFT, pbchf.RHF):
    '''RKS class adapted for PBCs.

    This is a literal duplication of the molecular RKS class with some `mol`
    variables replaced by `cell`.
    '''

    get_vsap = mol_ks.get_vsap
    init_guess_by_vsap = mol_ks.init_guess_by_vsap
    get_veff = get_veff
    energy_elec = mol_ks.energy_elec
    gen_response = gen_response

    def __init__(self, cell, kpt=numpy.zeros(3), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald')):
        pbchf.RHF.__init__(self, cell, kpt, exxdiv=exxdiv)
        KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        pbchf.RHF.dump_flags(self, verbose)
        KohnShamDFT.dump_flags(self, verbose)
        return self

    def to_hf(self):
        '''Convert to RHF object.'''
        from pyscf.pbc import scf
        return self._transfer_attrs_(scf.RHF(self.cell, self.kpt))

    def multigrid_numint(self, mesh=None):
        '''Apply the MultiGrid algorithm for XC numerical integartion'''
        mf = self.copy()
        mf._numint = multigrid.MultiGridNumInt(self.cell)
        if mesh is not None:
            mf._numint.mesh = mesh
        return mf

    def Gradients(self):
        from pyscf.pbc.grad import rks
        from pyscf.pbc.dft.multigrid import MultiGridNumInt2
        if not isinstance(self._numint, MultiGridNumInt2):
            raise NotImplementedError('pbc-RKS must be computed with MultiGridNumInt2')
        return rks.Gradients(self)

    to_gpu = lib.to_gpu



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''Non-relativistic UKS analytical nuclear gradients'''


import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.grad import rks as rks_grad
from pyscf.grad import uhf as uhf_grad
from pyscf.dft import numint, gen_grid
from pyscf import __config__


def get_veff(ks_grad, mol=None, dm=None):
    '''
    First order derivative of DFT effective potential matrix (wrt electron coordinates)

    Args:
        ks_grad : grad.uhf.Gradients or grad.uks.Gradients object
    '''
    if mol is None: mol = ks_grad.mol
    if dm is None: dm = ks_grad.base.make_rdm1()
    t0 = (logger.process_clock(), logger.perf_counter())

    mf = ks_grad.base
    ni = mf._numint
    grids, nlcgrids = rks_grad._initialize_grids(ks_grad)

    ni = mf._numint
    mem_now = lib.current_memory()[0]
    max_memory = max(2000, ks_grad.max_memory*.9-mem_now)
    if ks_grad.grid_response:
        exc, vxc = get_vxc_full_response(ni, mol, grids, mf.xc, dm,
                                         max_memory=max_memory,
                                         verbose=ks_grad.verbose)
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = rks_grad.get_nlc_vxc_full_response(
                ni, mol, nlcgrids, xc, dm[0]+dm[1],
                max_memory=max_memory, verbose=ks_grad.verbose)
            exc += enlc
            vxc += vnlc
        logger.debug1(ks_grad, 'sum(grids response) %s', exc.sum(axis=0))
    else:
        exc, vxc = get_vxc(ni, mol, grids, mf.xc, dm,
                           max_memory=max_memory, verbose=ks_grad.verbose)
        if mf.do_nlc():
            if ni.libxc.is_nlc(mf.xc):
                xc = mf.xc
            else:
                xc = mf.nlc
            enlc, vnlc = rks_grad.get_nlc_vxc(
                ni, mol, nlcgrids, xc, dm[0]+dm[1],
                max_memory=max_memory, verbose=ks_grad.verbose)
            vxc += vnlc
    t0 = logger.timer(ks_grad, 'vxc', *t0)

    if not ni.libxc.is_hybrid_xc(mf.xc):
        vj = ks_grad.get_j(mol, dm)
        vxc += vj[0] + vj[1]
    else:
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, spin=mol.spin)
        vj, vk = ks_grad.get_jk(mol, dm)
        vk *= hyb
        if omega != 0:
            vk += ks_grad.get_k(mol, dm, omega=omega) * (alpha - hyb)
        vxc += vj[0] + vj[1] - vk

    return lib.tag_array(vxc, exc1_grid=exc)


def get_vxc(ni, mol, grids, xc_code, dms, relativity=0, hermi=1,
            max_memory=2000, verbose=None):
    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()

    vmat = numpy.zeros((2,3,nao,nao))
    if xctype == 'LDA':
        ao_deriv = 1
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            rho_a = make_rho(0, ao[0], mask, xctype)
            rho_b = make_rho(1, ao[0], mask, xctype)
            vxc = ni.eval_xc_eff(xc_code, (rho_a,rho_b), 1, xctype=xctype)[1]
            wv = weight * vxc[:,0]
            aow = numint._scale_ao(ao[0], wv[0])
            rks_grad._d1_dot_(vmat[0], mol, ao[1:4], aow, mask, ao_loc, True)
            aow = numint._scale_ao(ao[0], wv[1])
            rks_grad._d1_dot_(vmat[1], mol, ao[1:4], aow, mask, ao_loc, True)

    elif xctype == 'GGA':
        ao_deriv = 2
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            rho_a = make_rho(0, ao[:4], mask, xctype)
            rho_b = make_rho(1, ao[:4], mask, xctype)
            vxc = ni.eval_xc_eff(xc_code, (rho_a,rho_b), 1, xctype=xctype)[1]
            wv = weight * vxc
            wv[:,0] *= .5
            rks_grad._gga_grad_sum_(vmat[0], mol, ao, wv[0], mask, ao_loc)
            rks_grad._gga_grad_sum_(vmat[1], mol, ao, wv[1], mask, ao_loc)

    elif xctype == 'NLC':
        raise NotImplementedError('NLC')

    elif xctype == 'MGGA':
        ao_deriv = 2
        for ao, mask, weight, coords \
                in ni.block_loop(mol, grids, nao, ao_deriv, max_memory):
            rho_a = make_rho(0, ao[:10], mask, xctype)
            rho_b = make_rho(1, ao[:10], mask, xctype)
            vxc = ni.eval_xc_eff(xc_code, (rho_a,rho_b), 1, xctype=xctype)[1]
            wv = weight * vxc
            wv[:,0] *= .5
            wv[:,4] *= .5
            rks_grad._gga_grad_sum_(vmat[0], mol, ao, wv[0], mask, ao_loc)
            rks_grad._gga_grad_sum_(vmat[1], mol, ao, wv[1], mask, ao_loc)
            rks_grad._tau_grad_dot_(vmat[0], mol, ao, wv[0,4], mask, ao_loc, True)
            rks_grad._tau_grad_dot_(vmat[1], mol, ao, wv[1,4], mask, ao_loc, True)

    exc = numpy.zeros((mol.natm,3))
    # - sign because nabla_X = -nabla_x
    return exc, -vmat


def get_vxc_full_response(ni, mol, grids, xc_code, dms, relativity=0, hermi=1,
                          max_memory=2000, verbose=None):
    '''Full response including the response of the grids'''
    xctype = ni._xc_type(xc_code)
    make_rho, nset, nao = ni._gen_rho_evaluator(mol, dms, hermi, False, grids)
    ao_loc = mol.ao_loc_nr()
    aoslices = mol.aoslice_by_atom()

    excsum = 0
    vmat = numpy.zeros((2,3,nao,nao))
    if xctype == 'LDA':
        ao_deriv = 1
        for atm_id, (coords, weight, weight1) \
                in enumerate(rks_grad.grids_response_cc(grids)):
            sh0, sh1 = aoslices[atm_id][:2]
            mask = gen_grid.make_mask(mol, coords)
            ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                            cutoff=grids.cutoff)
            rho_a = make_rho(0, ao[0], mask, xctype)
            rho_b = make_rho(1, ao[0], mask, xctype)
            exc, vxc = ni.eval_xc_eff(xc_code, (rho_a,rho_b), 1, xctype=xctype)[:2]
            wv = weight * vxc[:,0]

            vtmp = numpy.zeros((3,nao,nao))
            aow = numint._scale_ao(ao[0], wv[0])
            rks_grad._d1_dot_(vtmp, mol, ao[1:4], aow, mask, ao_loc, True)
            vmat[0] += vtmp
            excsum += numpy.einsum('r,r,nxr->nx', exc, rho_a+rho_b, weight1)
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms[0]) * 2

            vtmp = numpy.zeros((3,nao,nao))
            aow = numint._scale_ao(ao[0], wv[1])
            rks_grad._d1_dot_(vtmp, mol, ao[1:4], aow, mask, ao_loc, True)
            vmat[1] += vtmp
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms[1]) * 2

    elif xctype == 'GGA':
        ao_deriv = 2
        for atm_id, (coords, weight, weight1) \
                in enumerate(rks_grad.grids_response_cc(grids)):
            sh0, sh1 = aoslices[atm_id][:2]
            mask = gen_grid.make_mask(mol, coords)
            ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                            cutoff=grids.cutoff)
            rho_a = make_rho(0, ao[:4], mask, xctype)
            rho_b = make_rho(1, ao[:4], mask, xctype)
            exc, vxc = ni.eval_xc_eff(xc_code, (rho_a,rho_b), 1, xctype=xctype)[:2]
            wv = weight * vxc
            wv[:,0] *= .5

            vtmp = numpy.zeros((3,nao,nao))
            rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[0], mask, ao_loc)
            vmat[0] += vtmp
            excsum += numpy.einsum('r,r,nxr->nx', exc, rho_a[0]+rho_b[0], weight1)
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms[0]) * 2

            vtmp = numpy.zeros((3,nao,nao))
            rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[1], mask, ao_loc)
            vmat[1] += vtmp
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms[1]) * 2

    elif xctype == 'NLC':
        raise NotImplementedError('NLC')

    elif xctype == 'MGGA':
        ao_deriv = 2
        for atm_id, (coords, weight, weight1) \
                in enumerate(rks_grad.grids_response_cc(grids)):
            sh0, sh1 = aoslices[atm_id][:2]
            mask = gen_grid.make_mask(mol, coords)
            ao = ni.eval_ao(mol, coords, deriv=ao_deriv, non0tab=mask,
                            cutoff=grids.cutoff)
            rho_a = make_rho(0, ao[:10], mask, xctype)
            rho_b = make_rho(1, ao[:10], mask, xctype)
            exc, vxc = ni.eval_xc_eff(xc_code, (rho_a,rho_b), 1, xctype=xctype)[:2]
            wv = weight * vxc
            wv[:,0] *= .5
            wv[:,4] *= .5

            vtmp = numpy.zeros((3,nao,nao))
            rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[0], mask, ao_loc)
            rks_grad._tau_grad_dot_(vtmp, mol, ao, wv[0,4], mask, ao_loc, True)
            vmat[0] += vtmp
            excsum += numpy.einsum('r,r,nxr->nx', exc, rho_a[0]+rho_b[0], weight1)
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms[0]) * 2

            vtmp = numpy.zeros((3,nao,nao))
            rks_grad._gga_grad_sum_(vtmp, mol, ao, wv[1], mask, ao_loc)
            rks_grad._tau_grad_dot_(vtmp, mol, ao, wv[1,4], mask, ao_loc, True)
            vmat[1] += vtmp
            excsum[atm_id] += numpy.einsum('xij,ji->x', vtmp, dms[1]) * 2

    # - sign because nabla_X = -nabla_x
    return excsum, -vmat


class Gradients(uhf_grad.Gradients):

    grid_response = getattr(__config__, 'grad_uks_Gradients_grid_response', False)

    _keys = {'grid_response', 'grids', 'nlcgrids'}

    def __init__(self, mf):
        uhf_grad.Gradients.__init__(self, mf)
        self.grids = None
        self.nlcgrids = None

    def dump_flags(self, verbose=None):
        uhf_grad.Gradients.dump_flags(self, verbose)
        logger.info(self, 'grid_response = %s', self.grid_response)
        return self

    get_veff = get_veff

    def extra_force(self, atom_id, envs):
        '''Hook for extra contributions in analytical gradients.

        Contributions like the response of auxiliary basis in density fitting
        method, the grid response in DFT numerical integration can be put in
        this function.
        '''
        if self.grid_response:
            vhf = envs['vhf']
            log = envs['log']
            log.debug('grids response for atom %d %s',
                      atom_id, vhf.exc1_grid[atom_id])
            return vhf.exc1_grid[atom_id]
        else:
            return 0

Grad = Gradients

from pyscf import dft
dft.uks.UKS.Gradients = dft.uks_symm.UKS.Gradients = lib.class_as_method(Gradients)



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
1-electron Spin-free X2C approximation
'''

from functools import reduce
import numpy
import scipy.linalg
from pyscf import lib
from pyscf import gto
from pyscf.lib import logger
from pyscf.scf import hf
from pyscf.scf import ghf
from pyscf.x2c import x2c
from pyscf.data import nist


def sfx2c1e(mf):
    '''Spin-free X2C.
    For the given SCF object, it updates the hcore constructor.  All integrals
    are computed in the real spherical GTO basis.

    Args:
        mf : an SCF object

    Returns:
        An SCF object

    Examples:

    >>> mol = gto.M(atom='H 0 0 0; F 0 0 1', basis='ccpvdz', verbose=0)
    >>> mf = scf.RHF(mol).sfx2c1e()
    >>> mf.scf()

    >>> import pyscf.x2c.sfx2c1e
    >>> mol.symmetry = 1
    >>> mol.build(0, 0)
    >>> mf = pyscf.x2c.sfx2c1e.sfx2c1e(scf.UHF(mol))
    >>> mf.scf()
    '''
    assert isinstance(mf, hf.SCF)

    if isinstance(mf, x2c._X2C_SCF):
        if mf.with_x2c is None:
            mf.with_x2c = SpinFreeX2CHelper(mf.mol)
            return mf
        elif not isinstance(mf.with_x2c, SpinFreeX2CHelper):
            # An object associated to x2c.SpinOrbitalX2CHelper
            raise NotImplementedError
        else:
            return mf

    return lib.set_class(SFX2C1E_SCF(mf), (SFX2C1E_SCF, mf.__class__))

sfx2c = sfx2c1e

class SFX2C1E_SCF(x2c._X2C_SCF):
    '''
    Attributes for spin-free X2C:
        with_x2c : X2C object
    '''

    __name_mixin__ = 'sfX2C1e'

    _keys = {'with_x2c'}

    def __init__(self, mf):
        self.__dict__.update(mf.__dict__)
        self.with_x2c = SpinFreeX2CHelper(mf.mol)

    def undo_x2c(self):
        '''Remove the X2C Mixin'''
        obj = lib.view(self, lib.drop_class(self.__class__, SFX2C1E_SCF))
        del obj.with_x2c
        return obj

    def get_hcore(self, mol=None):
        if self.with_x2c:
            hcore = self.with_x2c.get_hcore(mol)
            if isinstance(self, ghf.GHF):
                hcore = scipy.linalg.block_diag(hcore, hcore)
            return hcore
        else:
            return super(x2c._X2C_SCF, self).get_hcore(mol)

    def dip_moment(self, mol=None, dm=None, unit='Debye', verbose=logger.NOTE,
                   picture_change=True, **kwargs):
        r''' Dipole moment calculation with picture change correction

        Args:
             mol: an instance of :class:`Mole`
             dm : a 2D ndarrays density matrices

        Kwarg:
            picture_chang (bool) : Whether to compute the dipole moment with
            picture change correction.

        Return:
            A list: the dipole moment on x, y and z component
        '''
        if mol is None: mol = self.mol
        if dm is None: dm = self.make_rdm1()
        log = logger.new_logger(mol, verbose)

        if not (isinstance(dm, numpy.ndarray) and dm.ndim == 2):
            # UHF density matrices
            dm = dm[0] + dm[1]

        if isinstance(self, ghf.GHF):
            nao = mol.nao_nr()
            dm = dm[:nao,:nao] + dm[nao:,nao:]

        with mol.with_common_orig((0,0,0)):
            if picture_change:
                xmol = self.with_x2c.get_xmol()[0]
                nao = xmol.nao
                prp = xmol.intor_symmetric('int1e_sprsp').reshape(3,4,nao,nao)[:,3]
                c1 = 0.5/lib.param.LIGHT_SPEED
                ao_dip = self.with_x2c.picture_change(('int1e_r', prp*c1**2))
            else:
                ao_dip = mol.intor_symmetric('int1e_r')

        el_dip = numpy.einsum('xij,ji->x', ao_dip, dm).real

        charges = mol.atom_charges()
        coords  = mol.atom_coords()
        nucl_dip = numpy.einsum('i,ix->x', charges, coords)
        mol_dip = nucl_dip - el_dip

        if unit.upper() == 'DEBYE':
            mol_dip *= nist.AU2DEBYE
            log.note('Dipole moment(X, Y, Z, Debye): %8.5f, %8.5f, %8.5f', *mol_dip)
        else:
            log.note('Dipole moment(X, Y, Z, A.U.): %8.5f, %8.5f, %8.5f', *mol_dip)
        return mol_dip

    def _transfer_attrs_(self, dst):
        if self.with_x2c and not hasattr(dst, 'with_x2c'):
            logger.warn(self, 'Destination object of to_hf/to_ks method is not '
                        'an X2C object. Convert dst to X2C object.')
            dst = dst.sfx2c()
        return hf.SCF._transfer_attrs_(self, dst)

    def to_gpu(self):
        obj = self.undo_x2c().to_gpu().sfx2c1e()
        return lib.to_gpu(self, obj)


class SpinFreeX2CHelper(x2c.X2CHelperBase):
    '''1-component X2c (spin-free part only)
    '''
    def get_hcore(self, mol=None):
        '''1-component X2c Foldy-Wouthuysen (FW Hamiltonian  (spin-free part only)
        '''
        if mol is None: mol = self.mol
        if mol.has_ecp():
            raise NotImplementedError

        xmol, contr_coeff = self.get_xmol(mol)
        c = lib.param.LIGHT_SPEED
        assert ('1E' in self.approx.upper())
        t = xmol.intor_symmetric('int1e_kin')
        v = xmol.intor_symmetric('int1e_nuc')
        s = xmol.intor_symmetric('int1e_ovlp')
        w = xmol.intor_symmetric('int1e_pnucp')
        if 'get_xmat' in self.__dict__:
            # If the get_xmat method is overwritten by user, build the X
            # matrix with the external get_xmat method
            x = self.get_xmat(xmol)
            h1 = x2c._get_hcore_fw(t, v, w, s, x, c)

        elif 'ATOM' in self.approx.upper():
            atom_slices = xmol.offset_nr_by_atom()
            nao = xmol.nao_nr()
            x = numpy.zeros((nao,nao))
            for ia in range(xmol.natm):
                ish0, ish1, p0, p1 = atom_slices[ia]
                shls_slice = (ish0, ish1, ish0, ish1)
                t1 = xmol.intor('int1e_kin', shls_slice=shls_slice)
                s1 = xmol.intor('int1e_ovlp', shls_slice=shls_slice)
                with xmol.with_rinv_at_nucleus(ia):
                    z = -xmol.atom_charge(ia)
                    v1 = z * xmol.intor('int1e_rinv', shls_slice=shls_slice)
                    w1 = z * xmol.intor('int1e_prinvp', shls_slice=shls_slice)
                x[p0:p1,p0:p1] = x2c._x2c1e_xmatrix(t1, v1, w1, s1, c)
            h1 = x2c._get_hcore_fw(t, v, w, s, x, c)

        else:
            h1 = x2c._x2c1e_get_hcore(t, v, w, s, c)

        if self.basis is not None:
            s22 = xmol.intor_symmetric('int1e_ovlp')
            s21 = gto.intor_cross('int1e_ovlp', xmol, mol)
            c = lib.cho_solve(s22, s21)
            h1 = reduce(numpy.dot, (c.T, h1, c))
        if self.xuncontract and contr_coeff is not None:
            h1 = reduce(numpy.dot, (contr_coeff.T, h1, contr_coeff))
        return h1

    def picture_change(self, even_operator=(None, None), odd_operator=None):
        '''Picture change for even_operator + odd_operator

        even_operator has two terms at diagonal blocks
        [ v  0 ]
        [ 0  w ]

        odd_operator has the term at off-diagonal blocks
        [ 0    p ]
        [ p^T  0 ]

        v, w, and p can be strings (integral name) or matrices.
        '''
        mol = self.mol
        xmol, c = self.get_xmol(mol)
        pc_mat = self._picture_change(xmol, even_operator, odd_operator)

        if self.basis is not None:
            s22 = xmol.intor_symmetric('int1e_ovlp')
            s21 = gto.mole.intor_cross('int1e_ovlp', xmol, mol)
            c = lib.cho_solve(s22, s21)

        elif self.xuncontract:
            pass

        else:
            return pc_mat

        if pc_mat.ndim == 2:
            return lib.einsum('pi,pq,qj->ij', c, pc_mat, c)
        else:
            return lib.einsum('pi,xpq,qj->xij', c, pc_mat, c)

    def get_xmat(self, mol=None):
        if mol is None:
            xmol = self.get_xmol(mol)[0]
        else:
            xmol = mol
        c = lib.param.LIGHT_SPEED
        assert ('1E' in self.approx.upper())

        if 'ATOM' in self.approx.upper():
            atom_slices = xmol.offset_nr_by_atom()
            nao = xmol.nao_nr()
            x = numpy.zeros((nao,nao))
            for ia in range(xmol.natm):
                ish0, ish1, p0, p1 = atom_slices[ia]
                shls_slice = (ish0, ish1, ish0, ish1)
                t1 = xmol.intor('int1e_kin', shls_slice=shls_slice)
                s1 = xmol.intor('int1e_ovlp', shls_slice=shls_slice)
                with xmol.with_rinv_at_nucleus(ia):
                    z = -xmol.atom_charge(ia)
                    v1 = z * xmol.intor('int1e_rinv', shls_slice=shls_slice)
                    w1 = z * xmol.intor('int1e_prinvp', shls_slice=shls_slice)
                x[p0:p1,p0:p1] = x2c._x2c1e_xmatrix(t1, v1, w1, s1, c)
        else:
            t = xmol.intor_symmetric('int1e_kin')
            v = xmol.intor_symmetric('int1e_nuc')
            s = xmol.intor_symmetric('int1e_ovlp')
            w = xmol.intor_symmetric('int1e_pnucp')
            x = x2c._x2c1e_xmatrix(t, v, w, s, c)
        return x

    def _get_rmat(self, x=None):
        '''The matrix (in AO basis) that changes metric from NESC metric to NR metric'''
        xmol = self.get_xmol()[0]
        if x is None:
            x = self.get_xmat(xmol)
        c = lib.param.LIGHT_SPEED
        s = xmol.intor_symmetric('int1e_ovlp')
        t = xmol.intor_symmetric('int1e_kin')
        s1 = s + reduce(numpy.dot, (x.conj().T, t, x)) * (.5/c**2)
        return x2c._get_r(s, s1)

    def hcore_deriv_generator(self, mol=None, deriv=1):
        from pyscf.x2c import sfx2c1e_grad
        from pyscf.x2c import sfx2c1e_hess
        if deriv == 1:
            return sfx2c1e_grad.hcore_grad_generator(self, mol)
        elif deriv == 2:
            return sfx2c1e_hess.hcore_hess_generator(self, mol)
        else:
            raise NotImplementedError

SpinFreeX2C = SpinFreeX2CHelper


if __name__ == '__main__':
    mol = gto.Mole()
    mol.build(
        verbose = 0,
        atom = [["O" , (0. , 0.     , 0.)],
                [1   , (0. , -0.757 , 0.587)],
                [1   , (0. , 0.757  , 0.587)] ],
        basis = 'ccpvdz-dk',
    )

    method = hf.RHF(mol)
    enr = method.kernel()
    print('E(NR) = %.12g' % enr)

    method = sfx2c1e(hf.RHF(mol))
    esfx2c = method.kernel()
    print('E(SFX2C1E) = %.12g' % esfx2c)
    method.with_x2c.basis = 'unc-ccpvqz-dk'
    print('E(SFX2C1E) = %.12g' % method.kernel())
    method.with_x2c.approx = 'atom1e'
    print('E(SFX2C1E) = %.12g' % method.kernel())

    mf = method.density_fit().undo_x2c().run()
    print('E(DF-NR) = %.12g' % mf.e_tot)
